"""Tests for event_media.py"""

import datetime
from typing import Any, Awaitable, Callable, Dict
from unittest.mock import patch
import logging

import pytest
import aiohttp

from google_nest_sdm import diagnostics, google_nest_api
from google_nest_sdm.device import Device
from google_nest_sdm.event import EventMessage, EventToken
from google_nest_sdm.event_media import InMemoryEventMediaStore
from google_nest_sdm.transcoder import Transcoder
from google_nest_sdm.camera_traits import (
    CameraMotionTrait,
    CameraPersonTrait,
    CameraClipPreviewTrait,
    CameraEventImageTrait,
)
from google_nest_sdm.doorbell_traits import DoorbellChimeTrait


from .conftest import (
    DeviceHandler,
    EventCallback,
    Recorder,
    assert_diagnostics,
)


_LOGGER = logging.getLogger(__name__)


IMAGE_CAMERA_TRAITS = [
    CameraEventImageTrait.NAME,
    CameraMotionTrait.NAME,
    CameraPersonTrait.NAME,
    CameraPersonTrait.NAME,
]
IMAGE_MOTION_ONLY_CAMERA_TRAITS = [CameraEventImageTrait.NAME, CameraMotionTrait.NAME]
IMAGE_DOORBELL_TRAITS = [
    CameraEventImageTrait.NAME,
    DoorbellChimeTrait.NAME,
    CameraMotionTrait.NAME,
]
CLIP_CAMERA_TRAITS = [
    CameraClipPreviewTrait.NAME,
    CameraMotionTrait.NAME,
    CameraPersonTrait.NAME,
]
CLIP_DOORBELL_TRAITS = [
    CameraClipPreviewTrait.NAME,
    DoorbellChimeTrait.NAME,
    CameraMotionTrait.NAME,
]


@pytest.fixture(name="device_traits")
def mock_device_traits() -> list[str]:
    """Fixture for tests to setup default device traits"""
    return []


@pytest.fixture(name="device_id")
def mock_device_id(
    device_handler: DeviceHandler,
    device_traits: list[str],
    api_client: Callable[[], Awaitable[google_nest_api.GoogleNestAPI]],
) -> str:
    return device_handler.add_device(traits={k: {} for k in device_traits})


class MediaRouter:
    """Fixture for serving event media."""

    def __init__(self, app: aiohttp.web.Application, recorder: Recorder) -> None:
        self.app = app
        self.recorder = recorder
        self.api_responses: dict[str, list[aiohttp.web.Response]] = {}
        self.image_responses: dict[str, aiohttp.web.Response] = {}
        self.clip_responses: dict[str, aiohttp.web.Response] = {}
        self.token_cnt = -1
        self.app.add_routes(
            [
                aiohttp.web.post(r"/{device_id:.*}:executeCommand", self.api_handler),
                aiohttp.web.get(r"/image-url/{device_id:.*}", self.image_token_handler),
                aiohttp.web.get(
                    r"/clip-url/{device_id:.*}/token/{token:.*}", self.clip_handler
                ),
            ]
        )

    def image(self, device_id: str) -> None:
        """Prepare an image response"""
        self.token_cnt += 1
        token = f"g.{self.token_cnt}.eventToken"
        # Prepare API to generate image url + token
        if device_id not in self.api_responses:
            self.api_responses[device_id] = []
        self.api_responses[device_id].append(
            aiohttp.web.json_response(
                {
                    "results": {
                        "url": f"image-url/{device_id}",
                        "token": token,
                    },
                }
            )
        )
        # Prepare image serving
        self.image_responses[token] = aiohttp.web.Response(
            body=f"image-bytes-{token}".encode()
        )

    def clip(self, device_id: str) -> str:
        """Prepare an clip response and return the url."""
        self.token_cnt += 1
        token = f"g.{self.token_cnt}.eventToken"
        # Prepare ckip serving
        self.clip_responses[token] = aiohttp.web.Response(
            body=f"clip-bytes-{token}".encode()
        )
        return f"clip-url/{device_id}/token/{token}"

    async def api_handler(self, request: aiohttp.web.Request) -> aiohttp.web.Response:
        device_id = request.match_info["device_id"]
        assert device_id
        _LOGGER.debug("API handler %s", device_id)
        responses = self.api_responses[device_id]
        response = responses.pop(0)
        return response

    async def image_token_handler(
        self, request: aiohttp.web.Request
    ) -> aiohttp.web.Response:
        device_id = request.match_info["device_id"]
        assert device_id
        _LOGGER.debug("Image token handler %s - %s", device_id, self.image_responses)
        auth_header = request.headers["Authorization"]
        assert auth_header.startswith("Basic ")
        token = auth_header[6:]
        if (response := self.image_responses.pop(token)) is None:
            raise ValueError(f"No image response configured for token {token}")
        return response

    async def clip_handler(self, request: aiohttp.web.Request) -> aiohttp.web.Response:
        device_id = request.match_info["device_id"]
        token = request.match_info["token"]
        _LOGGER.debug("Clip handler %s", device_id)
        if (response := self.clip_responses.pop(token)) is None:
            raise ValueError(f"No image response configured for token {token}")
        return response


@pytest.fixture(name="media_router", autouse=True)
def mock_media_router(app: aiohttp.web.Application, recorder: Recorder) -> MediaRouter:
    """Fixture for preparing API and content serving calls."""
    return MediaRouter(app, recorder)


@pytest.fixture
def event_cache_size() -> int:
    """Fixture to configure the event cache size."""
    return 8


@pytest.fixture
def event_fetch() -> bool:
    """Fixture to configure the event cache size."""
    return True


@pytest.fixture(autouse=True)
def mock_event_media_manager(
    media_router: MediaRouter, device: Device, event_fetch: bool, event_cache_size: int
) -> None:
    """Fixture to prepare the event media manager for fetching media."""
    device.event_media_manager.cache_policy.fetch = event_fetch
    device.event_media_manager.cache_policy.event_cache_size = event_cache_size


@pytest.fixture(name="transcoder")
def mock_transcoder(device: Device) -> Transcoder:
    """Fixture that enables a transcoder."""
    fake_transcoder = Transcoder("/bin/echo", "")
    device.event_media_manager.cache_policy.transcoder = fake_transcoder
    return fake_transcoder


@pytest.fixture(name="device")
async def mock_device(
    device_id: str,
    api_client: Callable[[], Awaitable[google_nest_api.GoogleNestAPI]],
) -> Device:
    api = await api_client()
    devices = await api.async_get_devices()
    for device in devices: 
        if device.name == device_id:
            return device
    raise ValueError("Invalid test state, couldn't find device.")


@pytest.mark.parametrize(
    ("device_traits", "event_cache_size"), [(IMAGE_MOTION_ONLY_CAMERA_TRAITS, 10)]
)
async def test_event_manager_event_expiration(
    media_router: MediaRouter,
    device: Device,
    event_message: Callable[[Dict[str, Any]], Awaitable[EventMessage]],
) -> None:
    ts1 = datetime.datetime.now(tz=datetime.timezone.utc)
    media_router.image(device.name)
    await device.async_handle_event(
        await event_message(
            {
                "eventId": "0120ecc7-3b57-4eb4-9941-91609f189fb4",
                "timestamp": ts1.isoformat(timespec="seconds"),
                "resourceUpdate": {
                    "name": device.name,
                    "events": {
                        "sdm.devices.events.CameraMotion.Motion": {
                            "eventSessionId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                            "eventId": "FWWVQVUdGNUlTU2V4MGV2aTNXV...",
                        },
                    },
                },
                "userId": "AVPHwEuBfnPOnTqzVFT4IONX2Qqhu9EJ4ubO-bNnQ-yi",
            }
        )
    )
    ts2 = ts1 + datetime.timedelta(seconds=5)
    media_router.image(device.name)
    await device.async_handle_event(
        await event_message(
            {
                "eventId": "a94b2115-3b57-4eb4-8830-80519f188ec9",
                "timestamp": ts2.isoformat(timespec="seconds"),
                "resourceUpdate": {
                    "name": device.name,
                    "events": {
                        "sdm.devices.events.CameraMotion.Motion": {
                            "eventSessionId": "DgY5Y3VKaTZwR3o4Y19YbTVfMF...",
                            "eventId": "ABCZQRUdGNUlTU2V4MGV3bRZ23...",
                        },
                    },
                },
                "userId": "AVPHwEuBfnPOnTqzVFT4IONX2Qqhu9EJ4ubO-bNnQ-yi",
            }
        )
    )

    # Event is in the past and is expired
    ts3 = ts1 - datetime.timedelta(seconds=90)
    media_router.image(device.name)  # Ignored
    await device.async_handle_event(
        await event_message(
            {
                "eventId": "b83c2115-3b57-4eb4-8830-80519f167fa8",
                "timestamp": ts3.isoformat(timespec="seconds"),
                "resourceUpdate": {
                    "name": device.name,
                    "events": {
                        "sdm.devices.events.CameraMotion.Motion": {
                            "eventSessionId": "EkY5Y3VKaTZwR3o4Y19YbTVfMF...",
                            "eventId": "1234QRUdGNUlTU2V4MGV3bRZ23...",
                        },
                    },
                },
                "userId": "AVPHwEuBfnPOnTqzVFT4IONX2Qqhu9EJ4ubO-bNnQ-yi",
            }
        )
    )

    event_media_manager = device.event_media_manager
    assert len(list(await event_media_manager.async_image_sessions())) == 2


@pytest.mark.parametrize(("device_traits"), [(IMAGE_MOTION_ONLY_CAMERA_TRAITS)])
async def test_event_manager_cache_expiration(
    media_router: MediaRouter,
    device: Device,
    # event_media_store: InMemoryEventMediaStore,
    event_message: Callable[[Dict[str, Any]], Awaitable[EventMessage]],
) -> None:
    num_events = 10
    event_tokens = []
    for i in range(0, num_events):
        media_router.image(device.name)
        ts = datetime.datetime.now(tz=datetime.timezone.utc) + datetime.timedelta(
            seconds=i
        )
        await device.async_handle_event(
            await event_message(
                {
                    "eventId": f"0120ecc7-{i}",
                    "timestamp": ts.isoformat(timespec="seconds"),
                    "resourceUpdate": {
                        "name": device.name,
                        "events": {
                            "sdm.devices.events.CameraMotion.Motion": {
                                "eventSessionId": f"CjY5Y3VK..{i}...",
                                "eventId": f"FWWVQVU..{i}...",
                            },
                        },
                    },
                    "userId": "AVPHwEuBfnPOnTqzVFT4IONX2Qqhu9EJ4ubO-bNnQ-yi",
                }
            )
        )
        image_sessions = await device.event_media_manager.async_image_sessions()
        event_tokens.append(image_sessions[0].event_token)

    assert_diagnostics(
        diagnostics.get_diagnostics(),
        {
            "event_media": {
                "save_media_count": 10,
                "remove_media_count": 2,
            },
        },
    )
    assert_diagnostics(
        device.get_diagnostics(),
        {
            "command": {
                "fetch_image_count": 10,
                "sdm.devices.commands.CameraEventImage.GenerateImage_count": 10,
            },
            "data": {
                "name": "**REDACTED**",
                "parentRelations": [],
                "traits": {
                    "sdm.devices.traits.CameraEventImage": {},
                    "sdm.devices.traits.CameraMotion": {},
                },
                "type": "sdm.devices.types.device-type1",
            },
            "event_media": {
                "event": 10,
                "event.fetch": 10,
                "event.new": 10,
                "fetch_image": 10,
                "fetch_image.save": 10,
                "load_image_sessions": 10,
                "sdm.devices.events.CameraMotion.Motion_count": 10,
            },
        },
    )

    event_media_manager = device.event_media_manager
    # All old items are evicted from the cache
    assert len(list(await event_media_manager.async_image_sessions())) == 8

    # Old items are evicted from the media store
    assert await event_media_manager.get_media_from_token(event_tokens[0]) is None
    assert await event_media_manager.get_media_from_token(event_tokens[1]) is None
    for i in range(2, num_events):
        media = await event_media_manager.get_media_from_token(event_tokens[i])
        assert media
        assert media.contents == f"image-bytes-g.{i}.eventToken".encode()

    assert_diagnostics(
        diagnostics.get_diagnostics(),
        {
            "event_media": {
                "load_media_count": 8,
                "save_media_count": 10,
                "remove_media_count": 2,
            },
        },
    )


@pytest.mark.parametrize("device_traits", [IMAGE_CAMERA_TRAITS])
async def test_prefetch_image_failure_in_session(
    media_router: MediaRouter,
    device: Device,
    event_message: Callable[[Dict[str, Any]], Awaitable[EventMessage]],
) -> None:
    """Exercise case where one image within a session fails."""
    # Send one failure response, then 3 other valid responses. The cache size
    # is too small so we're exercising events dropping out of the cache.
    media_router.image(device.name)
    media_router.api_responses[device.name].append(aiohttp.web.Response(status=502))

    now = datetime.datetime.now(tz=datetime.timezone.utc)
    ts1 = now + datetime.timedelta(seconds=1)
    await device.async_handle_event(
        await event_message(
            {
                "eventId": "0120ecc7-1",
                "timestamp": ts1.isoformat(timespec="seconds"),
                "resourceUpdate": {
                    "name": device.name,
                    "events": {
                        "sdm.devices.events.CameraMotion.Motion": {
                            "eventSessionId": "CjY5Y.......",
                            "eventId": "FWWVQVU..1...",
                        },
                    },
                },
                "userId": "AVPHwEuBfnPOnTqzVFT4IONX2Qqhu9EJ4ubO-bNnQ-yi",
            }
        )
    )
    ts2 = now + datetime.timedelta(seconds=2)
    await device.async_handle_event(
        await event_message(
            {
                "eventId": "0120ecc7-2",
                "timestamp": ts2.isoformat(timespec="seconds"),
                "resourceUpdate": {
                    "name": device.name,
                    "events": {
                        "sdm.devices.events.CameraPerson.Person": {
                            "eventSessionId": "CjY5Y.......",
                            "eventId": "FWWVQVU..2...",
                        },
                    },
                },
                "userId": "AVPHwEuBfnPOnTqzVFT4IONX2Qqhu9EJ4ubO-bNnQ-yi",
            }
        )
    )

    event_media_manager = device.event_media_manager

    events = list(await event_media_manager.async_image_sessions())
    assert len(events) == 1
    event = events[0]
    event_token = EventToken.decode(event.event_token)
    assert event_token.event_session_id == "CjY5Y......."
    assert event_token.event_id == "FWWVQVU..1..."
    assert event.event_type == "sdm.devices.events.CameraMotion.Motion"
    assert event.timestamp.isoformat(timespec="seconds") == ts1.isoformat(
        timespec="seconds"
    )


@pytest.mark.parametrize(("device", "mock_event_media_manager"), [(None, None)])
async def test_multi_device_events(
    media_router: MediaRouter,
    device: Device | None,
    device_handler: DeviceHandler,
    api_client: Callable[[], Awaitable[google_nest_api.GoogleNestAPI]],
    event_message: Callable[[Dict[str, Any]], Awaitable[EventMessage]],
) -> None:
    device_id1 = device_handler.add_device(
        traits={
            "sdm.devices.traits.CameraEventImage": {},
            "sdm.devices.traits.CameraMotion": {},
        }
    )
    device_id2 = device_handler.add_device(
        traits={
            "sdm.devices.traits.CameraEventImage": {},
            "sdm.devices.traits.CameraMotion": {},
        }
    )
    api = await api_client()
    devices = await api.async_get_devices()
    assert len(devices) == 2
    device = devices[0]
    assert device
    assert device.name == device_id1
    device = devices[1]
    assert device
    assert device.name == device_id2

    # Use shared event store for all devices
    store = InMemoryEventMediaStore()
    devices[0].event_media_manager.cache_policy.store = store
    devices[0].event_media_manager.cache_policy.fetch = True
    devices[1].event_media_manager.cache_policy.store = store
    devices[1].event_media_manager.cache_policy.fetch = True

    # Each device has no sessions
    event_media_manager = devices[0].event_media_manager
    assert len(list(await event_media_manager.async_image_sessions())) == 0
    event_media_manager = devices[1].event_media_manager
    assert len(list(await event_media_manager.async_image_sessions())) == 0

    ts = datetime.datetime.now(tz=datetime.timezone.utc)
    media_router.image(device_id1)
    await devices[0].async_handle_event(
        await event_message(
            {
                "eventId": "0120ecc7-1",
                "timestamp": ts.isoformat(timespec="seconds"),
                "resourceUpdate": {
                    "name": device_id1,
                    "events": {
                        "sdm.devices.events.CameraMotion.Motion": {
                            "eventSessionId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                            "eventId": "FWWVQVU..1...",
                        },
                    },
                },
                "userId": "AVPHwEuBfnPOnTqzVFT4IONX2Qqhu9EJ4ubO-bNnQ-yi",
            }
        )
    )

    # First device has a single event
    event_media_manager = devices[0].event_media_manager
    assert len(list(await event_media_manager.async_image_sessions())) == 1
    event_media_manager = devices[1].event_media_manager
    assert len(list(await event_media_manager.async_image_sessions())) == 0

    media_router.image(device_id2)
    await devices[1].async_handle_event(
        await event_message(
            {
                "eventId": "0120ecc7-2",
                "timestamp": ts.isoformat(timespec="seconds"),
                "resourceUpdate": {
                    "name": device_id2,
                    "events": {
                        "sdm.devices.events.CameraMotion.Motion": {
                            "eventSessionId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                            "eventId": "FWWVQVU..2...",
                        },
                    },
                },
                "userId": "AVPHwEuBfnPOnTqzVFT4IONX2Qqhu9EJ4ubO-bNnQ-yi",
            }
        )
    )

    # Each device has a single event
    event_media_manager = devices[0].event_media_manager
    assert len(list(await event_media_manager.async_image_sessions())) == 1
    event_media_manager = devices[1].event_media_manager
    assert len(list(await event_media_manager.async_image_sessions())) == 1


@pytest.mark.parametrize("device_traits", [IMAGE_DOORBELL_TRAITS])
async def test_event_session_image(
    media_router: MediaRouter,
    device: Device,
    event_message: Callable[[Dict[str, Any]], Awaitable[EventMessage]],
) -> None:
    ts1 = datetime.datetime.now(tz=datetime.timezone.utc)
    media_router.image(device.name)
    await device.async_handle_event(
        await event_message(
            {
                "eventId": "0120ecc7-3b57-4eb4-9941-91609f189fb4",
                "timestamp": ts1.isoformat(timespec="seconds"),
                "resourceUpdate": {
                    "name": device.name,
                    "events": {
                        "sdm.devices.events.CameraMotion.Motion": {
                            "eventSessionId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                            "eventId": "FWWVQVUdGNUlTU2V4MGV2aTNXV...",
                        },
                    },
                },
                "userId": "AVPHwEuBfnPOnTqzVFT4IONX2Qqhu9EJ4ubO-bNnQ-yi",
            }
        )
    )
    ts2 = ts1 + datetime.timedelta(seconds=5)
    media_router.image(device.name)
    await device.async_handle_event(
        await event_message(
            {
                "eventId": "90120ecc7-3b57-4eb4-9941-91609f278ec3",
                "timestamp": ts2.isoformat(timespec="seconds"),
                "resourceUpdate": {
                    "name": device.name,
                    "events": {
                        "sdm.devices.events.DoorbellChime.Chime": {
                            "eventSessionId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                            "eventId": "ABCZQRUdGNUlTU2V4MGV3bRZ23...",
                        },
                    },
                },
                "userId": "AVPHwEuBfnPOnTqzVFT4IONX2Qqhu9EJ4ubO-bNnQ-yi",
            }
        )
    )

    event_media_manager = device.event_media_manager

    events = list(await event_media_manager.async_image_sessions())
    assert len(events) == 2
    event = events[0]
    event_token = EventToken.decode(event.event_token)
    assert event_token.event_session_id == "CjY5Y3VKaTZwR3o4Y19YbTVfMF..."
    assert event_token.event_id == "ABCZQRUdGNUlTU2V4MGV3bRZ23..."
    assert event.event_type == "sdm.devices.events.DoorbellChime.Chime"
    assert event.timestamp.isoformat(timespec="seconds") == ts2.isoformat(
        timespec="seconds"
    )

    media = await event_media_manager.get_media_from_token(event.event_token)
    assert media
    assert media.contents == b"image-bytes-g.1.eventToken"
    assert media.content_type == "image/jpeg"

    event = events[1]
    event_token = EventToken.decode(event.event_token)
    assert event_token.event_session_id == "CjY5Y3VKaTZwR3o4Y19YbTVfMF..."
    assert event_token.event_id == "FWWVQVUdGNUlTU2V4MGV2aTNXV..."
    assert event.event_type == "sdm.devices.events.CameraMotion.Motion"
    assert event.timestamp.isoformat(timespec="seconds") == ts1.isoformat(
        timespec="seconds"
    )

    media = await event_media_manager.get_media_from_token(event.event_token)
    assert media
    assert media.contents == b"image-bytes-g.0.eventToken"
    assert media.content_type == "image/jpeg"


@pytest.mark.parametrize("device_traits", [CLIP_DOORBELL_TRAITS])
async def test_event_session_clip_preview(
    media_router: MediaRouter,
    device: Device,
    event_message: Callable[[Dict[str, Any]], Awaitable[EventMessage]],
) -> None:
    callback = EventCallback()
    device.event_media_manager.set_update_callback(callback.async_handle_event)

    ts1 = datetime.datetime.now(tz=datetime.timezone.utc)
    url = media_router.clip(device.name)
    await device.async_handle_event(
        await event_message(
            {
                "eventId": "0120ecc7-3b57-4eb4-9941-91609f189fb4",
                "timestamp": ts1.isoformat(timespec="seconds"),
                "resourceUpdate": {
                    "name": device.name,
                    "events": {
                        "sdm.devices.events.CameraMotion.Motion": {
                            "eventSessionId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                            "eventId": "n:1",
                        },
                        "sdm.devices.events.CameraClipPreview.ClipPreview": {
                            "eventSessionId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                            "previewUrl": url,
                        },
                    },
                },
                "userId": "AVPHwEuBfnPOnTqzVFT4IONX2Qqhu9EJ4ubO-bNnQ-yi",
                "eventThreadId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                "resourcegroup": [
                    "enterprises/project-id1/devices/device-id1",
                ],
                "eventThreadState": "STARTED",
            }
        )
    )
    assert callback.invoked
    assert len(callback.messages) == 1
    assert callback.messages[0].event_sessions
    assert "CjY5Y3VKaTZwR3o4Y19YbTVfMF..." in callback.messages[0].event_sessions
    session = callback.messages[0].event_sessions["CjY5Y3VKaTZwR3o4Y19YbTVfMF..."]
    assert "sdm.devices.events.CameraMotion.Motion" in session
    assert "sdm.devices.events.CameraClipPreview.ClipPreview" in session

    assert_diagnostics(
        device.get_diagnostics().get("event_media", {}),
        {
            "event": 1,
            "event.fetch": 1,
            "event.new": 1,
            "event.notify": 1,
            "fetch_clip": 1,
            "fetch_clip.save": 1,
            "sdm.devices.events.CameraClipPreview.ClipPreview_count": 1,
            "sdm.devices.events.CameraMotion.Motion_count": 1,
        },
    )
    ts2 = ts1 + datetime.timedelta(seconds=5)
    await device.async_handle_event(
        await event_message(
            {
                "eventId": "0120ecc7-3b57-4eb4-9941-91609f189fb4",
                "timestamp": ts2.isoformat(timespec="seconds"),
                "resourceUpdate": {
                    "name": device.name,
                    "events": {
                        "sdm.devices.events.DoorbellChime.Chime": {
                            "eventSessionId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                            "eventId": "n:2",
                        },
                    },
                },
                "userId": "AVPHwEuBfnPOnTqzVFT4IONX2Qqhu9EJ4ubO-bNnQ-yi",
                "eventThreadId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                "resourcegroup": [
                    "enterprises/project-id1/devices/device-id1",
                ],
                "eventThreadState": "ENDED",
            }
        )
    )
    event_media_manager = device.event_media_manager

    # Update callback was invoked because there is a new event type
    assert_diagnostics(
        diagnostics.get_diagnostics(),
        {
            "event_media": {
                "save_media_count": 1,
            },
        },
    )
    assert_diagnostics(
        device.get_diagnostics().get("event_media", {}),
        {
            "event": 2,
            "event.fetch": 2,
            "event.new": 1,
            "event.notify": 2,
            "event.update": 1,
            "fetch_clip": 2,
            "fetch_clip.save": 1,
            "fetch_clip.skip": 1,
            "sdm.devices.events.CameraClipPreview.ClipPreview_count": 1,
            "sdm.devices.events.CameraMotion.Motion_count": 1,
            "sdm.devices.events.DoorbellChime.Chime_count": 1,
        },
    )
    # New event published where we already have media
    assert len(callback.messages) == 2
    assert callback.messages[1].event_sessions
    assert "CjY5Y3VKaTZwR3o4Y19YbTVfMF..." in callback.messages[1].event_sessions
    session = callback.messages[1].event_sessions["CjY5Y3VKaTZwR3o4Y19YbTVfMF..."]
    assert "sdm.devices.events.DoorbellChime.Chime" in session

    events = list(await event_media_manager.async_clip_preview_sessions())
    assert len(events) == 1
    event = events[0]
    event_token = EventToken.decode(event.event_token)
    assert event_token.event_session_id == "CjY5Y3VKaTZwR3o4Y19YbTVfMF..."
    assert event_token.event_id == "n:1"
    assert event.event_types == [
        "sdm.devices.events.CameraMotion.Motion",
        "sdm.devices.events.DoorbellChime.Chime",
    ]
    assert event.timestamp.isoformat(timespec="seconds") == ts1.isoformat(
        timespec="seconds"
    )

    media = await event_media_manager.get_media_from_token(event.event_token)
    assert media
    assert media.contents == b"clip-bytes-g.0.eventToken"
    assert media.content_type == "video/mp4"


@pytest.mark.parametrize("device_traits", [CLIP_DOORBELL_TRAITS])
async def test_event_session_without_clip(
    media_router: MediaRouter,
    device: Device,
    event_message: Callable[[Dict[str, Any]], Awaitable[EventMessage]],
) -> None:
    callback = EventCallback()
    device.event_media_manager.set_update_callback(callback.async_handle_event)

    ts1 = datetime.datetime.now(tz=datetime.timezone.utc)
    await device.async_handle_event(
        await event_message(
            {
                "eventId": "0120ecc7-3b57-4eb4-9941-91609f189fb4",
                "timestamp": ts1.isoformat(timespec="seconds"),
                "resourceUpdate": {
                    "name": device.name,
                    "events": {
                        "sdm.devices.events.CameraMotion.Motion": {
                            "eventSessionId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                            "eventId": "n:1",
                        },
                    },
                },
                "userId": "AVPHwEuBfnPOnTqzVFT4IONX2Qqhu9EJ4ubO-bNnQ-yi",
                "eventThreadId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                "resourcegroup": [
                    "enterprises/project-id1/devices/device-id1",
                ],
                "eventThreadState": "STARTED",
            }
        )
    )

    # Event is received, but does not contain media
    # Do not invoke callback without media
    assert_diagnostics(
        device.get_diagnostics().get("event_media", {}),
        {
            "event": 1,
            "event.fetch": 1,
            "event.new": 1,
            "fetch_clip": 1,
            "fetch_clip.skip": 1,
            "sdm.devices.events.CameraMotion.Motion_count": 1,
        },
    )
    assert not callback.invoked

    ts2 = ts1 + datetime.timedelta(seconds=5)
    await device.async_handle_event(
        await event_message(
            {
                "eventId": "0120ecc7-3b57-4eb4-9941-91609f189fb4",
                "timestamp": ts2.isoformat(timespec="seconds"),
                "resourceUpdate": {
                    "name": device.name,
                    "events": {
                        "sdm.devices.events.DoorbellChime.Chime": {
                            "eventSessionId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                            "eventId": "n:2",
                        },
                    },
                },
                "userId": "AVPHwEuBfnPOnTqzVFT4IONX2Qqhu9EJ4ubO-bNnQ-yi",
                "eventThreadId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                "resourcegroup": [
                    "enterprises/project-id1/devices/device-id1",
                ],
                "eventThreadState": "ENDED",
            }
        )
    )

    # The event ENDED without any media, so notify
    assert callback.invoked  # type: ignore[unreachable]
    assert len(callback.messages) == 1  # type: ignore[unreachable]
    assert callback.messages[0].event_sessions
    assert "CjY5Y3VKaTZwR3o4Y19YbTVfMF..." in callback.messages[0].event_sessions
    session = callback.messages[0].event_sessions["CjY5Y3VKaTZwR3o4Y19YbTVfMF..."]
    assert "sdm.devices.events.DoorbellChime.Chime" in session
    assert "sdm.devices.events.CameraMotion.Motion" in session

    event_media_manager = device.event_media_manager

    # There are no event media clips
    events = list(await event_media_manager.async_clip_preview_sessions())
    assert not events


@pytest.mark.parametrize("device_traits", [CLIP_DOORBELL_TRAITS])
async def test_event_session_clip_preview_in_second_message(
    media_router: MediaRouter,
    device: Device,
    event_message: Callable[[Dict[str, Any]], Awaitable[EventMessage]],
) -> None:
    callback = EventCallback()
    device.event_media_manager.set_update_callback(callback.async_handle_event)

    ts1 = datetime.datetime.now(tz=datetime.timezone.utc)
    await device.async_handle_event(
        await event_message(
            {
                "eventId": "0120ecc7-3b57-4eb4-9941-91609f189fb4",
                "timestamp": ts1.isoformat(timespec="seconds"),
                "resourceUpdate": {
                    "name": device.name,
                    "events": {
                        "sdm.devices.events.DoorbellChime.Chime": {
                            "eventSessionId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                            "eventId": "n:1",
                        },
                    },
                },
                "userId": "AVPHwEuBfnPOnTqzVFT4IONX2Qqhu9EJ4ubO-bNnQ-yi",
                "eventThreadId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                "resourcegroup": [
                    "enterprises/project-id1/devices/device-id1",
                ],
                "eventThreadState": "STARTED",
            }
        )
    )

    # Do not invoke callback without media
    assert not callback.invoked
    assert_diagnostics(
        device.get_diagnostics().get("event_media", {}),
        {
            "event": 1,
            "event.fetch": 1,
            "event.new": 1,
            "fetch_clip": 1,
            "fetch_clip.skip": 1,
            "sdm.devices.events.DoorbellChime.Chime_count": 1,
        },
    )

    ts2 = ts1 + datetime.timedelta(seconds=5)
    clip_url = media_router.clip(device.name)
    await device.async_handle_event(
        await event_message(
            {
                "eventId": "0120ecc7-3b57-4eb4-9941-91609f189fb4",
                "timestamp": ts2.isoformat(timespec="seconds"),
                "resourceUpdate": {
                    "name": device.name,
                    "events": {
                        "sdm.devices.events.CameraClipPreview.ClipPreview": {
                            "eventSessionId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                            "previewUrl": clip_url,
                        },
                    },
                },
                "userId": "AVPHwEuBfnPOnTqzVFT4IONX2Qqhu9EJ4ubO-bNnQ-yi",
                "eventThreadId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                "resourcegroup": [
                    "enterprises/project-id1/devices/device-id1",
                ],
                "eventThreadState": "ENDED",
            }
        )
    )

    # Callback invoked now that media arrived. The previous events are now received.
    assert callback.invoked  # type: ignore[unreachable]
    assert len(callback.messages) == 1  # type: ignore[unreachable]
    assert callback.messages[0].event_sessions
    assert "CjY5Y3VKaTZwR3o4Y19YbTVfMF..." in callback.messages[0].event_sessions
    session = callback.messages[0].event_sessions["CjY5Y3VKaTZwR3o4Y19YbTVfMF..."]
    assert "sdm.devices.events.CameraClipPreview.ClipPreview" in session
    assert "sdm.devices.events.DoorbellChime.Chime" in session

    assert_diagnostics(  # type: ignore[unreachable]
        device.get_diagnostics().get("event_media", {}),
        {
            "event": 2,
            "event.fetch": 2,
            "event.new": 1,
            "event.update": 1,
            "event.notify": 1,
            "fetch_clip": 2,
            "fetch_clip.skip": 1,
            "fetch_clip.save": 1,
            "sdm.devices.events.CameraClipPreview.ClipPreview_count": 1,
            "sdm.devices.events.DoorbellChime.Chime_count": 1,
        },
    )

    event_media_manager = device.event_media_manager

    events = list(await event_media_manager.async_clip_preview_sessions())
    assert len(events) == 1
    event = events[0]
    event_token = EventToken.decode(event.event_token)
    assert event_token.event_session_id == "CjY5Y3VKaTZwR3o4Y19YbTVfMF..."
    assert event_token.event_id == "n:1"
    assert event.event_types == [
        "sdm.devices.events.DoorbellChime.Chime",
    ]
    assert event.timestamp.isoformat(timespec="seconds") == ts1.isoformat(
        timespec="seconds"
    )

    media = await event_media_manager.get_media_from_token(event.event_token)
    assert media
    assert media.contents == b"clip-bytes-g.0.eventToken"
    assert media.content_type == "video/mp4"


@pytest.mark.parametrize("device_traits", [CLIP_DOORBELL_TRAITS])
async def test_event_session_clip_preview_issue(
    media_router: MediaRouter,
    device: Device,
    event_message: Callable[[Dict[str, Any]], Awaitable[EventMessage]],
) -> None:
    """Exercise event session clip preview event ordering behavior.

    Reproduces issue https://github.com/home-assistant/core/issues/86314
    """
    callback = EventCallback()
    device.event_media_manager.set_update_callback(callback.async_handle_event)

    ts1 = datetime.datetime.now(tz=datetime.timezone.utc)
    await device.async_handle_event(
        await event_message(
            {
                "eventId": "0120ecc7-3b57-4eb4-9941-91609f189fb4",
                "timestamp": ts1.isoformat(timespec="seconds"),
                "resourceUpdate": {
                    "name": device.name,
                    "events": {
                        "sdm.devices.events.DoorbellChime.Chime": {
                            "eventSessionId": "1635497756",
                            "eventId": "n:1",
                        },
                    },
                },
                "userId": "AVPHwEuBfnPOnTqzVFT4IONX2Qqhu9EJ4ubO-bNnQ-yi",
                "eventThreadId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                "resourcegroup": [
                    "enterprises/project-id1/devices/device-id1",
                ],
                "eventThreadState": "STARTED",
            }
        )
    )

    assert_diagnostics(
        device.get_diagnostics().get("event_media", {}),
        {
            "event": 1,
            "event.fetch": 1,
            "event.new": 1,
            "fetch_clip": 1,
            "fetch_clip.skip": 1,
            "sdm.devices.events.DoorbellChime.Chime_count": 1,
        },
    )

    ts2 = ts1 + datetime.timedelta(seconds=5)
    clip_url = media_router.clip(device.name)
    await device.async_handle_event(
        await event_message(
            {
                "eventId": "0120ecc7-3b57-4eb4-9941-91609f189fb4",
                "timestamp": ts2.isoformat(timespec="seconds"),
                "resourceUpdate": {
                    "name": device.name,
                    "events": {
                        "sdm.devices.events.DoorbellChime.Chime": {
                            "eventSessionId": "1635497756",
                            "eventId": "n:1",
                        },
                        "sdm.devices.events.CameraClipPreview.ClipPreview": {
                            "eventSessionId": "1635497756",
                            "previewUrl": clip_url,
                        },
                    },
                },
                "userId": "AVPHwEuBfnPOnTqzVFT4IONX2Qqhu9EJ4ubO-bNnQ-yi",
                "eventThreadId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                "resourcegroup": [
                    "enterprises/project-id1/devices/device-id1",
                ],
                "eventThreadState": "UPDATED",
            }
        )
    )

    # Callback invoked now that media arrived
    assert callback.invoked  # type: ignore[unreachable]
    assert_diagnostics(  # type: ignore[unreachable]
        device.get_diagnostics().get("event_media", {}),
        {
            "event": 2,
            "event.fetch": 2,
            "event.new": 1,
            "event.update": 1,
            "event.notify": 1,
            "fetch_clip": 2,
            "fetch_clip.skip": 1,
            "fetch_clip.save": 1,
            "sdm.devices.events.CameraClipPreview.ClipPreview_count": 1,
            "sdm.devices.events.DoorbellChime.Chime_count": 2,
        },
    )
    # Event containing the media is delivered
    assert len(callback.messages) == 1
    assert callback.messages[0].event_sessions
    assert "1635497756" in callback.messages[0].event_sessions
    message = callback.messages[0].event_sessions["1635497756"]
    assert "sdm.devices.events.CameraClipPreview.ClipPreview" in message
    assert "sdm.devices.events.DoorbellChime.Chime" in message


@pytest.mark.parametrize("device_traits", [IMAGE_DOORBELL_TRAITS])
async def test_persisted_storage_image_event_media_keys(device: Device) -> None:
    data = {
        device.name: [
            {
                "event_session_id": "AVPHwEtyzgSxu6EuaIOfvz...",
                "events": {
                    "sdm.devices.events.CameraMotion.Motion": {
                        "event_type": "sdm.devices.events.CameraMotion.Motion",
                        "event_data": {
                            "eventSessionId": "AVPHwEtyzgSxu6EuaIOfvz...",
                            "eventId": "CiUA2vuxrwjZjb0daCbmE...",
                        },
                        "timestamp": "2021-12-23T06:35:35.791000+00:00",
                        "event_image_type": "image/jpeg",
                    },
                    "sdm.devices.events.DoorbellChime.Chime": {
                        "event_type": "sdm.devices.events.DoorbellChime.Chime",
                        "event_data": {
                            "eventSessionId": "AVPHwEtyzgSxu6EuaIOfvz...",
                            "eventId": "CiUA2vuxr_zoChpekrBmo...",
                        },
                        "timestamp": "2021-12-23T06:35:36.101000+00:00",
                        "event_image_type": "image/jpeg",
                    },
                },
                "event_media_keys": {
                    "CiUA2vuxrwjZjb0daCbmE...": (
                        "AVPHwEtyzgSxu6EuaIOfvzmr7-CiUA2vuxrwjZjb0daCbmE-motion.jpg"
                    ),
                    "CiUA2vuxr_zoChpekrBmo...": (
                        "AVPHwEtyzgSxu6EuaIOfvzmr7-CiUA2vuxr_zoChpekrBmo-doorbell.jpg"
                    ),
                },
            },
        ],
    }
    event_media_manager = device.event_media_manager
    store = event_media_manager.cache_policy.store
    await store.async_save(data)
    await store.async_save_media(
        "AVPHwEtyzgSxu6EuaIOfvzmr7-CiUA2vuxrwjZjb0daCbmE-motion.jpg",
        b"image-bytes-1",
    )
    await store.async_save_media(
        "AVPHwEtyzgSxu6EuaIOfvzmr7-CiUA2vuxr_zoChpekrBmo-doorbell.jpg",
        b"image-bytes-2",
    )

    event_media_manager = device.event_media_manager

    events = list(await event_media_manager.async_image_sessions())
    assert len(events) == 2
    event = events[0]
    event_token = EventToken.decode(event.event_token)
    assert event_token.event_session_id == "AVPHwEtyzgSxu6EuaIOfvz..."
    assert event_token.event_id == "CiUA2vuxr_zoChpekrBmo..."
    assert event.event_type == "sdm.devices.events.DoorbellChime.Chime"
    assert event.timestamp.isoformat(timespec="seconds") == "2021-12-23T06:35:36+00:00"

    media = await event_media_manager.get_media_from_token(event.event_token)
    assert media
    assert media.contents == b"image-bytes-2"
    assert media.content_type == "image/jpeg"

    event = events[1]
    event_token = EventToken.decode(event.event_token)
    assert event_token.event_session_id == "AVPHwEtyzgSxu6EuaIOfvz..."
    assert event_token.event_id == "CiUA2vuxrwjZjb0daCbmE..."
    assert event.event_type == "sdm.devices.events.CameraMotion.Motion"
    assert event.timestamp.isoformat(timespec="seconds") == "2021-12-23T06:35:35+00:00"

    media = await event_media_manager.get_media_from_token(event.event_token)
    assert media
    assert media.contents == b"image-bytes-1"
    assert media.content_type == "image/jpeg"

    # Test fallback to other media within the same session
    await store.async_remove_media(
        "AVPHwEtyzgSxu6EuaIOfvzmr7-CiUA2vuxr_zoChpekrBmo-doorbell.jpg"
    )
    assert await event_media_manager.get_media_from_token(event.event_token)
    await store.async_remove_media(
        "AVPHwEtyzgSxu6EuaIOfvzmr7-CiUA2vuxrwjZjb0daCbmE-motion.jpg",
    )
    assert not await event_media_manager.get_media_from_token(event.event_token)


@pytest.mark.parametrize("device_traits", [IMAGE_DOORBELL_TRAITS])
async def test_persisted_storage_image(device: Device) -> None:
    data = {
        device.name: [
            {
                "event_session_id": "AVPHwEtyzgSxu6EuaIOfvz...",
                "events": {
                    "sdm.devices.events.CameraMotion.Motion": {
                        "event_type": "sdm.devices.events.CameraMotion.Motion",
                        "event_data": {
                            "eventSessionId": "AVPHwEtyzgSxu6EuaIOfvz...",
                            "eventId": "CiUA2vuxrwjZjb0daCbmE...",
                        },
                        "timestamp": "2021-12-23T06:35:35.791000+00:00",
                        "event_image_type": "image/jpeg",
                    },
                    "sdm.devices.events.DoorbellChime.Chime": {
                        "event_type": "sdm.devices.events.DoorbellChime.Chime",
                        "event_data": {
                            "eventSessionId": "AVPHwEtyzgSxu6EuaIOfvz...",
                            "eventId": "CiUA2vuxr_zoChpekrBmo...",
                        },
                        "timestamp": "2021-12-23T06:35:36.101000+00:00",
                        "event_image_type": "image/jpeg",
                    },
                },
                "media_key": (
                    "AVPHwEtyzgSxu6EuaIOfvzmr7oaxdtpvXrJCJXcjIwQ4RQ6CMZW97Gb2dupC4uHJcx_NrAPRAPyD7KFraR32we-LAFgMjA-doorbell_chime.jpg"
                ),
            },
        ],
    }
    event_media_manager = device.event_media_manager
    store = event_media_manager.cache_policy.store
    await store.async_save(data)
    # Legacy storage where only one media is stored for multiple image events
    await store.async_save_media(
        "AVPHwEtyzgSxu6EuaIOfvzmr7oaxdtpvXrJCJXcjIwQ4RQ6CMZW97Gb2dupC4uHJcx_NrAPRAPyD7KFraR32we-LAFgMjA-doorbell_chime.jpg",
        b"image-bytes-1",
    )

    event_media_manager = device.event_media_manager

    events = list(await event_media_manager.async_image_sessions())
    assert len(events) == 2
    event = events[0]
    event_token = EventToken.decode(event.event_token)
    assert event_token.event_session_id == "AVPHwEtyzgSxu6EuaIOfvz..."
    assert event_token.event_id == "CiUA2vuxr_zoChpekrBmo..."
    assert event.event_type == "sdm.devices.events.DoorbellChime.Chime"
    assert event.timestamp.isoformat(timespec="seconds") == "2021-12-23T06:35:36+00:00"

    media = await event_media_manager.get_media_from_token(event.event_token)
    assert media
    assert media.contents == b"image-bytes-1"
    assert media.content_type == "image/jpeg"

    event = events[1]
    event_token = EventToken.decode(event.event_token)
    assert event_token.event_session_id == "AVPHwEtyzgSxu6EuaIOfvz..."
    assert event_token.event_id == "CiUA2vuxrwjZjb0daCbmE..."
    assert event.event_type == "sdm.devices.events.CameraMotion.Motion"
    assert event.timestamp.isoformat(timespec="seconds") == "2021-12-23T06:35:35+00:00"

    media = await event_media_manager.get_media_from_token(event.event_token)
    assert media
    assert media.contents == b"image-bytes-1"
    assert media.content_type == "image/jpeg"

    # Test failure where media key points to removed media
    await store.async_remove_media(
        "AVPHwEtyzgSxu6EuaIOfvzmr7oaxdtpvXrJCJXcjIwQ4RQ6CMZW97Gb2dupC4uHJcx_NrAPRAPyD7KFraR32we-LAFgMjA-doorbell_chime.jpg"
    )
    assert not await event_media_manager.get_media_from_token(event.event_token)


@pytest.mark.parametrize("device_traits", [CLIP_DOORBELL_TRAITS])
async def test_persisted_storage_clip_preview(device: Device) -> None:
    data = {
        device.name: [
            {
                "event_session_id": "1632710204",
                "events": {
                    "sdm.devices.events.CameraMotion.Motion": {
                        "event_type": "sdm.devices.events.CameraMotion.Motion",
                        "event_data": {
                            "eventSessionId": "1632710204",
                            "eventId": "n:1",
                        },
                        "timestamp": "2021-12-21T21:13:18.734000+00:00",
                        "event_image_type": "video/mp4",
                    },
                    "sdm.devices.events.DoorbellChime.Chime": {
                        "event_type": "sdm.devices.events.DoorbellChime.Chime",
                        "event_data": {
                            "eventSessionId": "1632710204",
                            "eventId": "n:2",
                        },
                        "timestamp": "2021-12-21T21:13:18.734000+00:00",
                        "event_image_type": "video/mp4",
                    },
                    "sdm.devices.events.CameraClipPreview.ClipPreview": {
                        "event_type": (
                            "sdm.devices.events.CameraClipPreview.ClipPreview"
                        ),
                        "event_data": {
                            "eventSessionId": "1632710204",
                            "previewUrl": "https://127.0.0.1/example",
                        },
                        "timestamp": "2021-12-21T21:13:18.734000+00:00",
                        "event_image_type": "video/mp4",
                    },
                },
                "media_key": "1640121198-1632710204-doorbell_chime.mp4",
            }
        ],
    }
    event_media_manager = device.event_media_manager
    store = event_media_manager.cache_policy.store
    await store.async_save(data)
    await store.async_save_media(
        "1640121198-1632710204-doorbell_chime.mp4", b"image-bytes-1"
    )

    events = list(await event_media_manager.async_clip_preview_sessions())
    assert len(events) == 1
    event = events[0]
    event_token = EventToken.decode(event.event_token)
    assert event_token.event_session_id == "1632710204"
    assert event_token.event_id == "n:1"
    assert event.event_types == [
        "sdm.devices.events.CameraMotion.Motion",
        "sdm.devices.events.DoorbellChime.Chime",
    ]
    assert event.timestamp.isoformat(timespec="seconds") == "2021-12-21T21:13:18+00:00"

    media = await event_media_manager.get_media_from_token(event.event_token)
    assert media
    assert media.contents == b"image-bytes-1"
    assert media.content_type == "video/mp4"

    # Test failure where media key points to removed media
    await store.async_remove_media("1640121198-1632710204-doorbell_chime.mp4")
    assert not await event_media_manager.get_media_from_token(event.event_token)


@pytest.mark.parametrize("device_traits", [IMAGE_DOORBELL_TRAITS])
async def test_event_image_lookup_failure(
    device: Device,
    event_message: Callable[[Dict[str, Any]], Awaitable[EventMessage]],
) -> None:
    # Disable pre-fetch
    event_media_manager = device.event_media_manager
    event_media_manager.cache_policy.fetch = False

    token = EventToken(
        "CjY5Y3VKaTZwR3o4Y19YbTVfMF...", "FWWVQVUdGNUlTU2V4MGV2aTNXV..."
    ).encode()
    assert not await event_media_manager.get_media_from_token(token)

    now = datetime.datetime.now(tz=datetime.timezone.utc)
    await device.async_handle_event(
        await event_message(
            {
                "eventId": "0120ecc7-3b57-4eb4-9941-91609f189fb4",
                "timestamp": now.isoformat(timespec="seconds"),
                "resourceUpdate": {
                    "name": device.name,
                    "events": {
                        "sdm.devices.events.CameraMotion.Motion": {
                            "eventSessionId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                            "eventId": "FWWVQVUdGNUlTU2V4MGV2aTNXV...",
                        },
                    },
                },
                "userId": "AVPHwEuBfnPOnTqzVFT4IONX2Qqhu9EJ4ubO-bNnQ-yi",
            }
        )
    )

    assert not await event_media_manager.get_media_from_token(token)


@pytest.mark.parametrize("device_traits", [CLIP_DOORBELL_TRAITS])
async def test_clip_preview_lookup_failure(
    device: Device,
    event_message: Callable[[Dict[str, Any]], Awaitable[EventMessage]],
) -> None:
    event_media_manager = device.event_media_manager
    # No media fetch so media is not visible
    event_media_manager.cache_policy.fetch = False

    token = EventToken("CjY5Y3VKaTZwR3o4Y19YbTVfMF...", "ignored-event-id").encode()
    assert not await event_media_manager.get_media_from_token(token)

    now = datetime.datetime.now(tz=datetime.timezone.utc)
    await device.async_handle_event(
        await event_message(
            {
                "eventId": "0120ecc7-3b57-4eb4-9941-91609f189fb4",
                "timestamp": now.isoformat(timespec="seconds"),
                "resourceUpdate": {
                    "name": device.name,
                    "events": {
                        "sdm.devices.events.CameraMotion.Motion": {
                            "eventSessionId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                            "eventId": "n:1",
                        },
                        "sdm.devices.events.CameraClipPreview.ClipPreview": {
                            "eventSessionId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                            "previewUrl": "image-url-1",
                        },
                    },
                },
                "userId": "AVPHwEuBfnPOnTqzVFT4IONX2Qqhu9EJ4ubO-bNnQ-yi",
                "eventThreadId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                "resourcegroup": [
                    "enterprises/project-id1/devices/device-id1",
                ],
                "eventThreadState": "STARTED",
            }
        )
    )
    assert not await event_media_manager.get_media_from_token(token)


@pytest.mark.parametrize("device_traits", [CLIP_CAMERA_TRAITS])
async def test_clip_preview_transcode(
    media_router: MediaRouter,
    device: Device,
    event_message: Callable[[Dict[str, Any]], Awaitable[EventMessage]],
    transcoder: Transcoder,
) -> None:
    ts1 = datetime.datetime.now(tz=datetime.timezone.utc)
    clip_url = media_router.clip(device.name)
    await device.async_handle_event(
        await event_message(
            {
                "eventId": "0120ecc7-3b57-4eb4-9941-91609f189fb4",
                "timestamp": ts1.isoformat(timespec="seconds"),
                "resourceUpdate": {
                    "name": device.name,
                    "events": {
                        "sdm.devices.events.CameraMotion.Motion": {
                            "eventSessionId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF",
                            "eventId": "n:1",
                        },
                        "sdm.devices.events.CameraClipPreview.ClipPreview": {
                            "eventSessionId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF",
                            "previewUrl": clip_url,
                        },
                    },
                },
                "userId": "AVPHwEuBfnPOnTqzVFT4IONX2Qqhu9EJ4ubO-bNnQ-yi",
                "eventThreadId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF",
                "resourcegroup": [
                    "enterprises/project-id1/devices/device-id1",
                ],
                "eventThreadState": "STARTED",
            }
        )
    )

    event_media_manager = device.event_media_manager

    events = list(await event_media_manager.async_clip_preview_sessions())
    assert len(events) == 1
    event = events[0]
    event_token = EventToken.decode(event.event_token)
    assert event_token.event_session_id == "CjY5Y3VKaTZwR3o4Y19YbTVfMF"
    assert event_token.event_id == "n:1"
    assert event.event_types == [
        "sdm.devices.events.CameraMotion.Motion",
    ]
    assert event.timestamp.isoformat(timespec="seconds") == ts1.isoformat(
        timespec="seconds"
    )

    cnt = 0

    def values() -> Callable[[str], bool]:
        def func(filename: str) -> bool:
            nonlocal cnt
            cnt = cnt + 1
            if cnt == 1:
                # input file exists
                return True
            return False

        return func

    with patch("google_nest_sdm.transcoder.os.path.exists", new_callable=values), patch(
        "google_nest_sdm.event_media.InMemoryEventMediaStore.async_load_media",
        return_value=b"fake-video-thumb-bytes",
    ):
        media = await event_media_manager.get_clip_thumbnail_from_token(
            event.event_token
        )
        assert media
        assert media.contents == b"fake-video-thumb-bytes"
        assert media.content_type == "image/gif"

    # Test cache
    with patch(
        "google_nest_sdm.event_media.InMemoryEventMediaStore.async_load_media",
        return_value=b"fake-video-thumb-bytes",
    ):
        media = await event_media_manager.get_clip_thumbnail_from_token(
            event.event_token
        )
        assert media
        assert media.contents == b"fake-video-thumb-bytes"
        assert media.content_type == "image/gif"

    ts2 = ts1 + datetime.timedelta(seconds=5)
    clip_url = media_router.clip(device.name)
    await device.async_handle_event(
        await event_message(
            {
                "eventId": "0120ecc7-3b57-4eb4-9941-91609f189fb4",
                "timestamp": ts2.isoformat(timespec="seconds"),
                "resourceUpdate": {
                    "name": device.name,
                    "events": {
                        "sdm.devices.events.CameraMotion.Motion": {
                            "eventSessionId": "DjY5Y3VKaTZwR3o4Y19YbTVfMF",
                            "eventId": "n:1",
                        },
                        "sdm.devices.events.CameraClipPreview.ClipPreview": {
                            "eventSessionId": "DjY5Y3VKaTZwR3o4Y19YbTVfMF",
                            "previewUrl": clip_url,
                        },
                    },
                },
                "userId": "AVPHwEuBfnPOnTqzVFT4IONX2Qqhu9EJ4ubO-bNnQ-yi",
                "eventThreadId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF",
                "resourcegroup": [
                    "enterprises/project-id1/devices/device-id1",
                ],
                "eventThreadState": "STARTED",
            }
        )
    )

    events = list(await event_media_manager.async_clip_preview_sessions())
    assert len(events) == 2
    event = events[0]
    event_token = EventToken.decode(event.event_token)
    assert event_token.event_session_id == "DjY5Y3VKaTZwR3o4Y19YbTVfMF"

    # Test failure mode
    with patch("google_nest_sdm.transcoder.os.path.exists", return_value=False):
        assert not await event_media_manager.get_clip_thumbnail_from_token(
            event.event_token
        )

    with patch(
        "google_nest_sdm.event_media.InMemoryEventMediaStore.async_load_media",
        return_value=None,
    ):
        assert not await event_media_manager.get_clip_thumbnail_from_token(
            event.event_token
        )

    # Disable transcoding
    event_media_manager.cache_policy.transcoder = None
    assert not await event_media_manager.get_clip_thumbnail_from_token(
        event.event_token
    )


@pytest.mark.parametrize(
    ("device_traits", "event_cache_size"), [(CLIP_CAMERA_TRAITS, 5)]
)
async def test_event_manager_event_expiration_with_transcode(
    media_router: MediaRouter,
    device: Device,
    event_message: Callable[[Dict[str, Any]], Awaitable[EventMessage]],
    transcoder: Transcoder,
) -> None:
    event_media_manager = device.event_media_manager

    num_events = 7
    event_tokens = []
    for i in range(0, num_events):
        ts = datetime.datetime.now(tz=datetime.timezone.utc) + datetime.timedelta(
            seconds=i
        )
        clip = media_router.clip(device.name)
        await device.async_handle_event(
            await event_message(
                {
                    "eventId": f"0120ecc7-{i}",
                    "timestamp": ts.isoformat(timespec="seconds"),
                    "resourceUpdate": {
                        "name": device.name,
                        "events": {
                            "sdm.devices.events.CameraMotion.Motion": {
                                "eventSessionId": f"CjY5Y3VK..{i}...",
                                "eventId": f"n:{i}",
                            },
                            "sdm.devices.events.CameraClipPreview.ClipPreview": {
                                "eventSessionId": f"CjY5Y3VK..{i}...",
                                "previewUrl": clip,
                            },
                        },
                    },
                    "userId": "AVPHwEuBfnPOnTqzVFT4IONX2Qqhu9EJ4ubO-bNnQ-yi",
                }
            )
        )
        events = list(await event_media_manager.async_clip_preview_sessions())
        assert len(events) > 0
        event_token = events[0].event_token
        media = await event_media_manager.get_media_from_token(event_token)
        assert media
        assert media.contents == f"clip-bytes-g.{i}.eventToken".encode()

        event_tokens.append(event_token)

        cnt = 0

        def values() -> Callable[[str], bool]:
            def func(filename: str) -> bool:
                nonlocal cnt
                cnt = cnt + 1
                if cnt == 1:
                    # input file exists
                    return True
                return False

            return func

        # Cache a clip thumbnail to ensure it is expired later
        with patch(
            "google_nest_sdm.transcoder.os.path.exists", new_callable=values
        ), patch(
            "google_nest_sdm.event_media.InMemoryEventMediaStore.async_load_media",
            return_value=b"fake-video-thumb-bytes",
        ):
            media = await event_media_manager.get_clip_thumbnail_from_token(event_token)
            assert media
            assert media.contents == b"fake-video-thumb-bytes"
            assert media.content_type == "image/gif"

    event_media_manager = device.event_media_manager
    # All old items are evicted from the cache
    assert len(list(await event_media_manager.async_clip_preview_sessions())) == 5

    # Old items are evicted from the media store
    assert await event_media_manager.get_media_from_token(event_tokens[0]) is None
    assert await event_media_manager.get_media_from_token(event_tokens[1]) is None
    for token in event_tokens[2:]:
        assert await event_media_manager.get_media_from_token(token) is not None


@pytest.mark.parametrize("device_traits", [IMAGE_DOORBELL_TRAITS])
async def test_unsupported_event_trait(
    media_router: MediaRouter,
    device: Device,
    event_message: Callable[[Dict[str, Any]], Awaitable[EventMessage]],
) -> None:
    ts = datetime.datetime.now(tz=datetime.timezone.utc)
    clip_url = media_router.clip(device.name)
    await device.async_handle_event(
        await event_message(
            {
                "eventId": "0120ecc7-3b57-4eb4-9941-91609f189fb4",
                "timestamp": ts.isoformat(timespec="seconds"),
                "resourceUpdate": {
                    "name": device.name,
                    "events": {
                        "sdm.devices.events.CameraClipPreview.ClipPreview": {
                            "eventSessionId": "CjY5Y3VK..0...",
                            "previewUrl": clip_url,
                        },
                    },
                },
                "userId": "AVPHwEuBfnPOnTqzVFT4IONX2Qqhu9EJ4ubO-bNnQ-yi",
            }
        )
    )
    event_media_manager = device.event_media_manager
    assert len(list(await event_media_manager.async_image_sessions())) == 0
    assert len(list(await event_media_manager.async_clip_preview_sessions())) == 0


@pytest.mark.parametrize("device_traits", [IMAGE_DOORBELL_TRAITS])
async def test_unknown_event_type(device: Device) -> None:
    data = {
        device.name: [
            {
                "event_session_id": "AVPHwEtyzgSxu6EuaIOfvz...",
                "events": {
                    "sdm.devices.events.Ignored.Ignored": {
                        "event_type": "sdm.devices.events.Ignored.Ignored",
                        "event_data": {
                            "eventSessionId": "AVPHwEtyzgSxu6EuaIOfvz...",
                            "eventId": "CiUA2vuxrwjZjb0daCbmE...",
                        },
                        "timestamp": "2021-12-23T06:35:35.791000+00:00",
                        "event_image_type": "image/jpeg",
                    },
                    "sdm.devices.events.DoorbellChime.Chime": {
                        "event_type": "sdm.devices.events.DoorbellChime.Chime",
                        "event_data": {
                            "eventSessionId": "AVPHwEtyzgSxu6EuaIOfvz...",
                            "eventId": "CiUA2vuxr_zoChpekrBmo...",
                        },
                        "timestamp": "2021-12-23T06:35:36.101000+00:00",
                        "event_image_type": "image/jpeg",
                    },
                },
                "event_media_keys": {
                    "CiUA2vuxrwjZjb0daCbmE...": (
                        "AVPHwEtyzgSxu6EuaIOfvzmr7-CiUA2vuxrwjZjb0daCbmE-motion.jpg"
                    ),
                    "CiUA2vuxr_zoChpekrBmo...": (
                        "AVPHwEtyzgSxu6EuaIOfvzmr7-CiUA2vuxr_zoChpekrBmo-doorbell.jpg"
                    ),
                },
            },
        ],
    }
    event_media_manager = device.event_media_manager
    store = event_media_manager.cache_policy.store
    await store.async_save(data)
    await store.async_save_media(
        "AVPHwEtyzgSxu6EuaIOfvzmr7-CiUA2vuxrwjZjb0daCbmE-motion.jpg",
        b"image-bytes-1",
    )
    await store.async_save_media(
        "AVPHwEtyzgSxu6EuaIOfvzmr7-CiUA2vuxr_zoChpekrBmo-doorbell.jpg",
        b"image-bytes-2",
    )

    event_media_manager = device.event_media_manager

    events = list(await event_media_manager.async_image_sessions())
    assert len(events) == 1
    event = events[0]
    event_token = EventToken.decode(event.event_token)
    assert event_token.event_session_id == "AVPHwEtyzgSxu6EuaIOfvz..."
    assert event_token.event_id == "CiUA2vuxr_zoChpekrBmo..."
    assert event.event_type == "sdm.devices.events.DoorbellChime.Chime"
    assert event.timestamp.isoformat(timespec="seconds") == "2021-12-23T06:35:36+00:00"



@pytest.mark.parametrize("device_traits", [IMAGE_DOORBELL_TRAITS])
async def test_invalid_events_persisted(device: Device) -> None:
    data = {
        device.name: [
            {
                "event_session_id": "AVPHwEtyzgSxu6EuaIOfvz...",
                "events": {
                    "sdm.devices.events.CameraMotion.Motion": {
                        "event_type": "sdm.devices.events.CameraMotion.Motion",
                    },
                    "sdm.devices.events.DoorbellChime.Chime": {
                        "event_type": "sdm.devices.events.DoorbellChime.Chime",
                        "event_data": {
                            "eventSessionId": "AVPHwEtyzgSxu6EuaIOfvz...",
                            "eventId": "CiUA2vuxr_zoChpekrBmo...",
                        },
                        "timestamp": "2021-12-23T06:35:36.101000+00:00",
                        "event_image_type": "image/jpeg",
                    },
                },
                "event_media_keys": {
                    "CiUA2vuxr_zoChpekrBmo...": (
                        "AVPHwEtyzgSxu6EuaIOfvzmr7-CiUA2vuxr_zoChpekrBmo-doorbell.jpg"
                    ),
                },
            },
        ],
    }
    event_media_manager = device.event_media_manager
    store = event_media_manager.cache_policy.store
    await store.async_save(data)
    await store.async_save_media(
        "AVPHwEtyzgSxu6EuaIOfvzmr7-CiUA2vuxrwjZjb0daCbmE-motion.jpg",
        b"image-bytes-1",
    )
    await store.async_save_media(
        "AVPHwEtyzgSxu6EuaIOfvzmr7-CiUA2vuxr_zoChpekrBmo-doorbell.jpg",
        b"image-bytes-2",
    )

    event_media_manager = device.event_media_manager

    with pytest.raises(ValueError, match="EventMediaModelItem has invalid value"):
        list(await event_media_manager.async_image_sessions())



"""Test for camera traits."""

import datetime
from typing import Any, Awaitable, Callable, Dict

import aiohttp
import pytest

from google_nest_sdm import google_nest_api
from google_nest_sdm.camera_traits import EventImageType, StreamingProtocol
from google_nest_sdm.device import Device

from .conftest import (
    DeviceHandler,
    NewHandler,
    NewImageHandler,
    Recorder,
    assert_diagnostics,
)


def test_camera_image_traits(fake_device: Callable[[Dict[str, Any]], Device]) -> None:
    raw = {
        "name": "my/device/name",
        "traits": {
            "sdm.devices.traits.CameraImage": {
                "maxImageResolution": {
                    "width": 500,
                    "height": 300,
                }
            },
        },
    }
    device = fake_device(raw)
    assert "sdm.devices.traits.CameraImage" in device.traits
    trait = device.traits["sdm.devices.traits.CameraImage"]
    assert trait.max_image_resolution.width == 500
    assert trait.max_image_resolution.height == 300


@pytest.mark.parametrize(
    "data",
    [
        ({}),
        ({"maxImageResolution": {}}),
        ({"maxImageResolution": {"width": 1024}}),
        ({"maxImageResolution": {"height": 1024}}),
    ],
)
def test_otional_fields(
    fake_device: Callable[[Dict[str, Any]], Device], data: dict[str, Any]
) -> None:
    raw = {
        "name": "my/device/name",
        "traits": {
            "sdm.devices.traits.CameraImage": data,
        },
    }
    device = fake_device(raw)
    assert "sdm.devices.traits.CameraImage" in device.traits
    assert device.camera_image


def test_camera_live_stream_traits(
    fake_device: Callable[[Dict[str, Any]], Device]
) -> None:
    raw = {
        "name": "my/device/name",
        "traits": {
            "sdm.devices.traits.CameraLiveStream": {
                "maxVideoResolution": {
                    "width": 500,
                    "height": 300,
                },
                "videoCodecs": ["H264"],
                "audioCodecs": ["AAC"],
            },
        },
    }
    device = fake_device(raw)
    assert "sdm.devices.traits.CameraLiveStream" in device.traits
    trait = device.traits["sdm.devices.traits.CameraLiveStream"]
    assert trait.max_video_resolution.width == 500
    assert trait.max_video_resolution.height == 300
    assert trait.video_codecs == ["H264"]
    assert trait.audio_codecs == ["AAC"]
    # Default value
    assert trait.supported_protocols == [StreamingProtocol.RTSP]


def test_camera_live_stream_webrtc_protocol(
    fake_device: Callable[[Dict[str, Any]], Device]
) -> None:
    raw = {
        "name": "my/device/name",
        "traits": {
            "sdm.devices.traits.CameraLiveStream": {
                "supportedProtocols": ["WEB_RTC"],
            },
        },
    }
    device = fake_device(raw)
    assert "sdm.devices.traits.CameraLiveStream" in device.traits
    trait = device.traits["sdm.devices.traits.CameraLiveStream"]
    assert trait.supported_protocols == [StreamingProtocol.WEB_RTC]


def test_camera_live_stream_multiple_protocols(
    fake_device: Callable[[Dict[str, Any]], Device]
) -> None:
    raw = {
        "name": "my/device/name",
        "traits": {
            "sdm.devices.traits.CameraLiveStream": {
                "supportedProtocols": ["WEB_RTC", "RTSP"],
            },
        },
    }
    device = fake_device(raw)
    assert "sdm.devices.traits.CameraLiveStream" in device.traits
    trait = device.traits["sdm.devices.traits.CameraLiveStream"]
    assert trait.supported_protocols == [
        StreamingProtocol.WEB_RTC,
        StreamingProtocol.RTSP,
    ]


def test_camera_live_stream_unknown_protocols(
    fake_device: Callable[[Dict[str, Any]], Device]
) -> None:
    raw = {
        "name": "my/device/name",
        "traits": {
            "sdm.devices.traits.CameraLiveStream": {
                "supportedProtocols": ["WEB_RTC", "XXX"],
            },
        },
    }
    device = fake_device(raw)
    assert "sdm.devices.traits.CameraLiveStream" in device.traits
    trait = device.traits["sdm.devices.traits.CameraLiveStream"]
    assert trait.supported_protocols == [StreamingProtocol.WEB_RTC]


@pytest.mark.parametrize(
    "data",
    [
        ({}),
        ({"maxVideoResolution": {}}),
        ({"maxVideoResolution": {"width": 1024}}),
        ({"maxVideoResolution": {"height": 1024}}),
        ({"videoCodecs": []}),
        ({"audioCodecs": []}),
    ],
)
def test_camera_live_stream_optional_fields(
    fake_device: Callable[[Dict[str, Any]], Device], data: dict[str, Any]
) -> None:
    raw = {
        "name": "my/device/name",
        "traits": {
            "sdm.devices.traits.CameraLiveStream": data,
        },
    }
    device = fake_device(raw)
    assert "sdm.devices.traits.CameraLiveStream" in device.traits
    assert device.camera_live_stream


@pytest.mark.parametrize(
    "trait",
    [
        "sdm.devices.traits.CameraMotion",
        "sdm.devices.traits.CameraPerson",
        "sdm.devices.traits.CameraSound",
        "sdm.devices.traits.CameraClipPreview",
        "sdm.devices.traits.CameraEventImage",
    ],
)
def test_image_event_traits(
    trait: str, fake_device: Callable[[Dict[str, Any]], Device]
) -> None:
    raw = {
        "name": "my/device/name",
        "traits": {
            trait: {},
        },
    }
    device = fake_device(raw)
    assert trait in device.traits


async def test_camera_live_stream_rtsp(
    app: aiohttp.web.Application,
    recorder: Recorder,
    device_handler: DeviceHandler,
    api_client: Callable[[], Awaitable[google_nest_api.GoogleNestAPI]],
) -> None:
    device_id = device_handler.add_device(
        traits={
            "sdm.devices.traits.CameraLiveStream": {
                "maxVideoResolution": {
                    "width": 500,
                    "height": 300,
                },
                "videoCodecs": ["H264"],
                "audioCodecs": ["AAC"],
            },
        }
    )

    post_handler = NewHandler(
        recorder,
        [
            {
                "results": {
                    "streamUrls": {
                        "rtspUrl": "rtsps://someurl.com/CjY5Y3VKaTfMF?auth=g.0.token"
                    },
                    "streamExtensionToken": "CjY5Y3VKaTfMF",
                    "streamToken": "g.0.token",
                    "expiresAt": "2018-01-04T18:30:00.000Z",
                },
            },
            {
                "results": {
                    "streamExtensionToken": "dGNUlTU2CjY5Y3VKaTZwR3o4Y1...",
                    "streamToken": "g.1.newStreamingToken",
                    "expiresAt": "2019-01-04T18:30:00.000Z",
                }
            },
            {
                "results": {
                    "streamExtensionToken": "last-token...",
                    "streamToken": "g.2.newStreamingToken",
                    "expiresAt": "2020-01-04T18:30:00.000Z",
                }
            },
            {},
        ],
    )
    app.router.add_post(f"/{device_id}:executeCommand", post_handler)

    api = await api_client()
    devices = await api.async_get_devices()
    assert len(devices) == 1
    device = devices[0]
    assert device.name == device_id
    trait = device.traits["sdm.devices.traits.CameraLiveStream"]
    stream = await trait.generate_rtsp_stream()
    assert recorder.request == {
        "command": "sdm.devices.commands.CameraLiveStream.GenerateRtspStream",
        "params": {},
    }
    assert stream.stream_token == "g.0.token"
    assert stream.expires_at == datetime.datetime(
        2018, 1, 4, 18, 30, tzinfo=datetime.timezone.utc
    )
    assert stream.rtsp_stream_url == "rtsps://someurl.com/CjY5Y3VKaTfMF?auth=g.0.token"

    stream = await stream.extend_rtsp_stream()
    assert recorder.request == {
        "command": "sdm.devices.commands.CameraLiveStream.ExtendRtspStream",
        "params": {
            "streamExtensionToken": "CjY5Y3VKaTfMF",
        },
    }
    assert stream.stream_token == "g.1.newStreamingToken"
    assert stream.expires_at == datetime.datetime(
        2019, 1, 4, 18, 30, tzinfo=datetime.timezone.utc
    )
    assert (
        stream.rtsp_stream_url
        == "rtsps://someurl.com/CjY5Y3VKaTfMF?auth=g.1.newStreamingToken"
    )

    stream = await stream.extend_rtsp_stream()
    assert recorder.request == {
        "command": "sdm.devices.commands.CameraLiveStream.ExtendRtspStream",
        "params": {
            "streamExtensionToken": "dGNUlTU2CjY5Y3VKaTZwR3o4Y1...",
        },
    }
    assert stream.stream_token == "g.2.newStreamingToken"
    assert stream.expires_at == datetime.datetime(
        2020, 1, 4, 18, 30, tzinfo=datetime.timezone.utc
    )
    assert (
        stream.rtsp_stream_url
        == "rtsps://someurl.com/CjY5Y3VKaTfMF?auth=g.2.newStreamingToken"
    )

    await stream.stop_rtsp_stream()
    assert recorder.request == {
        "command": "sdm.devices.commands.CameraLiveStream.StopRtspStream",
        "params": {
            "streamExtensionToken": "last-token...",
        },
    }

    assert_diagnostics(
        device.get_diagnostics(),
        {
            "data": {
                "name": "**REDACTED**",
                "parentRelations": [],
                "traits": {
                    "sdm.devices.traits.CameraLiveStream": {
                        "maxVideoResolution": {
                            "width": 500,
                            "height": 300,
                        },
                        "videoCodecs": ["H264"],
                        "audioCodecs": ["AAC"],
                        "supportedProtocols": ["RTSP"],
                    }
                },
                "type": "sdm.devices.types.device-type1",
            },
            "command": {
                "sdm.devices.commands.CameraLiveStream.ExtendRtspStream_count": 2,
                "sdm.devices.commands.CameraLiveStream.GenerateRtspStream_count": 1,
                "sdm.devices.commands.CameraLiveStream.StopRtspStream_count": 1,
            },
        },
    )


async def test_camera_live_stream_web_rtc(
    app: aiohttp.web.Application,
    recorder: Recorder,
    device_handler: DeviceHandler,
    api_client: Callable[[], Awaitable[google_nest_api.GoogleNestAPI]],
) -> None:
    device_id = device_handler.add_device(
        traits={
            "sdm.devices.traits.CameraLiveStream": {
                "supportedProtocols": ["WEB_RTC"],
            },
        }
    )

    post_handler = NewHandler(
        recorder,
        [
            {
                "results": {
                    "answerSdp": "some-answer",
                    "expiresAt": "2018-01-04T18:30:00.000Z",
                    "mediaSessionId": "JxdTxkkatHk4kVnXlKzQICbfVR...",
                },
            },
            {
                "results": {
                    "expiresAt": "2019-01-04T18:30:00.000Z",
                    "mediaSessionId": "JxdTxkkatHk4kVnXlKzQICbfVR...",
                }
            },
            {
                "results": {
                    "expiresAt": "2020-01-04T18:30:00.000Z",
                    "mediaSessionId": "JxdTxkkatHk4kVnXlKzQICbfVR...",
                }
            },
            {},
        ],
    )
    app.router.add_post(f"/{device_id}:executeCommand", post_handler)

    api = await api_client()
    devices = await api.async_get_devices()
    assert len(devices) == 1
    device = devices[0]
    assert device.name == device_id
    trait = device.traits["sdm.devices.traits.CameraLiveStream"]
    assert trait.supported_protocols == [StreamingProtocol.WEB_RTC]
    stream = await trait.generate_web_rtc_stream("a=recvonly")
    assert recorder.request == {
        "command": "sdm.devices.commands.CameraLiveStream.GenerateWebRtcStream",
        "params": {
            "offerSdp": "a=recvonly",
        },
    }
    assert stream.answer_sdp == "some-answer"
    assert stream.expires_at == datetime.datetime(
        2018, 1, 4, 18, 30, tzinfo=datetime.timezone.utc
    )
    assert stream.media_session_id == "JxdTxkkatHk4kVnXlKzQICbfVR..."

    stream = await stream.extend_stream()
    expected_request = {
        "command": "sdm.devices.commands.CameraLiveStream.ExtendWebRtcStream",
        "params": {
            "mediaSessionId": "JxdTxkkatHk4kVnXlKzQICbfVR...",
        },
    }
    assert expected_request == recorder.request
    assert "some-answer" == stream.answer_sdp
    assert (
        datetime.datetime(2019, 1, 4, 18, 30, tzinfo=datetime.timezone.utc)
        == stream.expires_at
    )
    assert "JxdTxkkatHk4kVnXlKzQICbfVR..." == stream.media_session_id

    stream = await stream.extend_stream()
    assert recorder.request == {
        "command": "sdm.devices.commands.CameraLiveStream.ExtendWebRtcStream",
        "params": {
            "mediaSessionId": "JxdTxkkatHk4kVnXlKzQICbfVR...",
        },
    }
    assert stream.answer_sdp == "some-answer"
    assert stream.expires_at == datetime.datetime(
        2020, 1, 4, 18, 30, tzinfo=datetime.timezone.utc
    )
    assert "JxdTxkkatHk4kVnXlKzQICbfVR..." == stream.media_session_id

    await stream.stop_stream()
    assert recorder.request == {
        "command": "sdm.devices.commands.CameraLiveStream.StopWebRtcStream",
        "params": {
            "mediaSessionId": "JxdTxkkatHk4kVnXlKzQICbfVR...",
        },
    }


async def test_camera_event_image(
    app: aiohttp.web.Application,
    recorder: Recorder,
    device_handler: DeviceHandler,
    api_client: Callable[[], Awaitable[google_nest_api.GoogleNestAPI]],
) -> None:
    device_id = device_handler.add_device(
        traits={"sdm.devices.traits.CameraEventImage": {}}
    )

    post_handler = NewHandler(
        recorder,
        [
            {
                "results": {
                    "url": "https://domain/sdm_event/dGNUlTU2CjY5Y3VKaTZwR3o4Y",
                    "token": "g.0.eventToken",
                },
            }
        ],
    )
    app.router.add_post(f"/{device_id}:executeCommand", post_handler)

    api = await api_client()
    devices = await api.async_get_devices()
    assert len(devices) == 1
    device = devices[0]
    assert device.name == device_id
    trait = device.traits["sdm.devices.traits.CameraEventImage"]
    image = await trait.generate_image("some-eventId")
    assert recorder.request == {
        "command": "sdm.devices.commands.CameraEventImage.GenerateImage",
        "params": {"eventId": "some-eventId"},
    }
    assert image.url == "https://domain/sdm_event/dGNUlTU2CjY5Y3VKaTZwR3o4Y"
    assert image.token == "g.0.eventToken"
    assert image.event_image_type == EventImageType.IMAGE


async def test_camera_event_image_bytes(
    app: aiohttp.web.Application,
    recorder: Recorder,
    device_handler: DeviceHandler,
    api_client: Callable[[], Awaitable[google_nest_api.GoogleNestAPI]],
) -> None:
    device_id = device_handler.add_device(
        traits={"sdm.devices.traits.CameraEventImage": {}}
    )

    post_handler = NewHandler(
        recorder,
        [
            {
                "results": {
                    "url": "image-url",
                    "token": "g.0.eventToken",
                },
            }
        ],
    )
    image_handler = NewImageHandler([b"image-bytes"], token="g.0.eventToken")

    app.router.add_post(f"/{device_id}:executeCommand", post_handler)
    app.router.add_get("/image-url", image_handler)

    api = await api_client()
    devices = await api.async_get_devices()
    assert len(devices) == 1
    device = devices[0]
    assert device.name == device_id
    trait = device.traits["sdm.devices.traits.CameraEventImage"]
    event_image = await trait.generate_image("some-eventId")
    image_bytes = await event_image.contents()
    assert image_bytes == b"image-bytes"



"""Tests for device properties."""

from typing import Any, Callable, Dict

import pytest

from google_nest_sdm.device import Device

from .conftest import assert_diagnostics


def test_device_id(fake_device: Callable[[Dict[str, Any]], Device]) -> None:
    device = fake_device(
        {
            "name": "my/device/name",
            "type": "sdm.devices.types.SomeDeviceType",
        }
    )
    assert "my/device/name" == device.name
    assert "sdm.devices.types.SomeDeviceType" == device.type


def test_no_traits(fake_device: Callable[[Dict[str, Any]], Device]) -> None:
    device = fake_device(
        {
            "name": "my/device/name",
        }
    )
    assert "my/device/name" == device.name
    assert "sdm.devices.traits.Info" not in device.traits


def test_empty_traits(fake_device: Callable[[Dict[str, Any]], Device]) -> None:
    device = fake_device(
        {
            "name": "my/device/name",
            "traits": {},
        }
    )
    assert "my/device/name" == device.name
    assert "sdm.devices.traits.Info" not in device.traits


def test_no_name(fake_device: Callable[[Dict[str, Any]], Device]) -> None:
    with pytest.raises(ValueError, match="'name' is required"):
        fake_device(
            {
                "traits": {},
            }
        )


def test_no_parent_relations(fake_device: Callable[[Dict[str, Any]], Device]) -> None:
    device = fake_device(
        {
            "name": "my/device/name",
        }
    )
    assert "my/device/name" == device.name
    assert {} == device.parent_relations


def test_empty_parent_relations(
    fake_device: Callable[[Dict[str, Any]], Device]
) -> None:
    device = fake_device(
        {
            "name": "my/device/name",
            "parentRelations": [],
        }
    )
    assert "my/device/name" == device.name
    assert {} == device.parent_relations


def test_invalid_parent_relations(
    fake_device: Callable[[Dict[str, Any]], Device]
) -> None:
    """Invalid parentRelations should be ignored."""
    device = fake_device(
        {
            "name": "my/device/name",
            "parentRelations": [{}],
        }
    )
    assert "my/device/name" == device.name
    assert {} == device.parent_relations


def test_parent_relation(fake_device: Callable[[Dict[str, Any]], Device]) -> None:
    device = fake_device(
        {
            "name": "my/device/name",
            "parentRelations": [
                {
                    "parent": "my/structure/or/room",
                    "displayName": "Some Name",
                },
            ],
        }
    )
    assert "my/device/name" == device.name
    assert {"my/structure/or/room": "Some Name"} == device.parent_relations

    assert_diagnostics(
        device.get_diagnostics(),
        {
            "data": {
                "name": "**REDACTED**",
                "parentRelations": [
                    {
                        "parent": "**REDACTED**",
                        "displayName": "**REDACTED**",
                    }
                ],
            },
        },
    )


def test_multiple_parent_relations(
    fake_device: Callable[[Dict[str, Any]], Device]
) -> None:
    device = fake_device(
        {
            "name": "my/device/name",
            "parentRelations": [
                {
                    "parent": "my/structure/or/room1",
                    "displayName": "Some Name1",
                },
                {
                    "parent": "my/structure/or/room2",
                    "displayName": "Some Name2",
                },
            ],
        }
    )
    assert "my/device/name" == device.name
    assert {
        "my/structure/or/room1": "Some Name1",
        "my/structure/or/room2": "Some Name2",
    } == device.parent_relations



"""Base library for all traits."""

from __future__ import annotations

from abc import ABC
from enum import StrEnum
from typing import Any, Mapping

import aiohttp
from mashumaro.types import SerializableType

from .auth import AbstractAuth
from .diagnostics import Diagnostics

DEVICE_TRAITS = "traits"
TRAITS = "traits"


class TraitType(StrEnum):
    """Traits for SDM devices."""

    CAMERA_IMAGE = "sdm.devices.traits.CameraImage"
    CAMERA_LIVE_STREAM = "sdm.devices.traits.CameraLiveStream"
    CAMERA_EVENT_IMAGE = "sdm.devices.traits.CameraEventImage"
    CAMERA_MOTION = "sdm.devices.traits.CameraMotion"
    CAMERA_PERSON = "sdm.devices.traits.CameraPerson"
    CAMERA_SOUND = "sdm.devices.traits.CameraSound"
    CAMERA_CLIP_PREVIEW = "sdm.devices.traits.CameraClipPreview"
    CONNECTIVITY = "sdm.devices.traits.Connectivity"
    FAN = "sdm.devices.traits.Fan"
    INFO = "sdm.devices.traits.Info"
    HUMIDITY = "sdm.devices.traits.Humidity"
    TEMPERATURE = "sdm.devices.traits.Temperature"
    DOORBELL_CHIME = "sdm.devices.traits.DoorbellChime"
    THERMOSTAT_ECO = "sdm.devices.traits.ThermostatEco"
    THERMOSTAT_HVAC = "sdm.devices.traits.ThermostatHvac"
    THERMOSTAT_MODE = "sdm.devices.traits.ThermostatMode"
    THERMOSTAT_TEMPERATURE_SETPOINT = "sdm.devices.traits.ThermostatTemperatureSetpoint"


class Command(SerializableType):
    """Base class for executing commands."""

    def __init__(self, device_id: str, auth: AbstractAuth, diagnostics: Diagnostics):
        """Initialize Command."""
        self._device_id = device_id
        self._auth = auth
        self._diagnostics = diagnostics

    async def execute(self, data: Mapping[str, Any]) -> aiohttp.ClientResponse:
        """Run the command."""
        assert self._auth
        cmd = data.get("command", "execute")
        with self._diagnostics.timer(cmd):
            return await self._auth.post(f"{self._device_id}:executeCommand", json=data)

    async def execute_json(self, data: Mapping[str, Any]) -> dict[str, Any]:
        """Run the command and return a json result."""
        assert self._auth
        cmd = data.get("command", "execute")
        with self._diagnostics.timer(cmd):
            return await self._auth.post_json(
                f"{self._device_id}:executeCommand", json=data
            )

    async def fetch_image(self, url: str, basic_auth: str | None = None) -> bytes:
        """Fetch an image at the specified url."""
        headers: dict[str, Any] = {}
        if basic_auth:
            headers = {"Authorization": f"Basic {basic_auth}"}
        with self._diagnostics.timer("fetch_image"):
            resp = await self._auth.get(url, headers=headers)
            return await resp.read()


class CommandDataClass(ABC):
    """Base model that supports commands."""

    def __post_init__(self) -> None:
        self._cmd: Command | None = None

    @property
    def cmd(self) -> Command:
        """Helper for executing commands, used internally by the trait"""
        if not self._cmd:
            raise ValueError("Device trait in invalid state")
        return self._cmd



import json
from typing import Awaitable, Callable
from unittest.mock import patch
import re
from http import HTTPStatus

import aiohttp
import pytest

from google_nest_sdm import google_nest_api
from google_nest_sdm.auth import AbstractAuth
from google_nest_sdm.exceptions import (
    ApiException,
    AuthException,
    NotFoundException,
    ApiForbiddenException,
)

from .conftest import (
    FAKE_TOKEN,
    PROJECT_ID,
    DeviceHandler,
    NewHandler,
    Recorder,
    StructureHandler,
    reply_handler,
)


async def test_get_device(
    device_handler: DeviceHandler,
    api_client: Callable[[], Awaitable[google_nest_api.GoogleNestAPI]],
) -> None:
    device_id = device_handler.add_device(device_type="sdm.devices.types.device-type")

    api = await api_client()

    device = await api.async_get_device(device_id.split("/")[-1])
    assert device
    assert device.name == device_id
    assert device.type == "sdm.devices.types.device-type"


async def test_get_devices(
    device_handler: DeviceHandler,
    api_client: Callable[[], Awaitable[google_nest_api.GoogleNestAPI]],
) -> None:
    device_id1 = device_handler.add_device(device_type="sdm.devices.types.device-type1")
    device_id2 = device_handler.add_device(device_type="sdm.devices.types.device-type2")

    api = await api_client()

    devices = await api.async_get_devices()
    assert len(devices) == 2
    assert devices[0].name == device_id1
    assert devices[0].type == "sdm.devices.types.device-type1"
    assert devices[1].name == device_id2
    assert devices[1].type == "sdm.devices.types.device-type2"


async def test_fan_set_timer(
    app: aiohttp.web.Application,
    recorder: Recorder,
    device_handler: DeviceHandler,
    api_client: Callable[[], Awaitable[google_nest_api.GoogleNestAPI]],
) -> None:
    device_id = device_handler.add_device(
        traits={
            "sdm.devices.traits.Fan": {
                "timerMode": "OFF",
            },
        }
    )
    post_handler = NewHandler(recorder, [{}])
    app.router.add_post(f"/{device_id}:executeCommand", post_handler)

    api = await api_client()
    devices = await api.async_get_devices()
    assert len(devices) == 1
    device = devices[0]
    assert device_id == device.name
    trait = device.traits["sdm.devices.traits.Fan"]
    assert trait.timer_mode == "OFF"
    await trait.set_timer("ON", 3600)
    assert recorder.request == {
        "command": "sdm.devices.commands.Fan.SetTimer",
        "params": {
            "timerMode": "ON",
            "duration": "3600s",
        },
    }


async def test_get_structure(
    app: aiohttp.web.Application,
    structure_handler: StructureHandler,
    api_client: Callable[[], Awaitable[google_nest_api.GoogleNestAPI]],
) -> None:
    structure_id = structure_handler.add_structure(
        traits={
            "sdm.structures.traits.Info": {
                "customName": "some-name",
            }
        }
    )

    api = await api_client()
    structure = await api.async_get_structure(structure_id.split("/")[-1])
    assert structure
    assert structure.name == structure_id
    assert "sdm.structures.traits.Info" in structure.traits


async def test_get_structures(
    app: aiohttp.web.Application,
    structure_handler: StructureHandler,
    api_client: Callable[[], Awaitable[google_nest_api.GoogleNestAPI]],
) -> None:
    structure_id1 = structure_handler.add_structure(
        traits={
            "sdm.structures.traits.Info": {
                "customName": "some-name1",
            }
        }
    )
    structure_id2 = structure_handler.add_structure(
        {
            "sdm.structures.traits.Info": {
                "customName": "some-name2",
            }
        }
    )

    api = await api_client()
    structures = await api.async_get_structures()
    assert len(structures) == 2
    assert structures[0].name == structure_id1
    assert "sdm.structures.traits.Info" in structures[0].traits
    assert structures[1].name == structure_id2
    assert "sdm.structures.traits.Info" in structures[1].traits


async def test_client_error(
    app: aiohttp.web.Application,
    api_client: Callable[[], Awaitable[google_nest_api.GoogleNestAPI]],
) -> None:
    # No server endpoint registered
    api = await api_client()
    with patch(
        "google_nest_sdm.google_nest_api.AbstractAuth._request",
        side_effect=aiohttp.ClientConnectionError(),
    ), pytest.raises(ApiException):
        await api.async_get_structures()


async def test_api_get_error(
    app: aiohttp.web.Application,
    api_client: Callable[[], Awaitable[google_nest_api.GoogleNestAPI]],
) -> None:
    # No server endpoint registered
    api = await api_client()
    with pytest.raises(ApiException):
        await api.async_get_structures()


async def test_api_post_error(
    app: aiohttp.web.Application,
    recorder: Recorder,
    device_handler: DeviceHandler,
    api_client: Callable[[], Awaitable[google_nest_api.GoogleNestAPI]],
) -> None:
    device_id = device_handler.add_device(
        traits={
            "sdm.devices.traits.ThermostatTemperatureSetpoint": {
                "heatCelsius": 23.0,
                "coolCelsius": 24.0,
            },
        }
    )

    async def fail_handler(request: aiohttp.web.Request) -> aiohttp.web.Response:
        assert request.headers["Authorization"] == "Bearer %s" % FAKE_TOKEN
        return aiohttp.web.Response(status=502)

    app.router.add_post(f"/{device_id}:executeCommand", fail_handler)

    api = await api_client()
    devices = await api.async_get_devices()
    assert len(devices) == 1
    device = devices[0]
    assert device.name == device_id
    trait = device.traits["sdm.devices.traits.ThermostatTemperatureSetpoint"]
    assert trait.heat_celsius == 23.0
    assert trait.cool_celsius == 24.0

    with pytest.raises(ApiException):
        await trait.set_heat(25.0)


async def test_auth_refresh(
    app: aiohttp.web.Application,
    recorder: Recorder,
    device_handler: DeviceHandler,
    refreshing_auth_client: Callable[[], Awaitable[AbstractAuth]],
) -> None:
    device_handler.token = "updated-token"
    device_id = device_handler.add_device(traits={})

    async def auth_handler(request: aiohttp.web.Request) -> aiohttp.web.Response:
        return aiohttp.web.json_response({"token": "updated-token"})

    app.router.add_get("/refresh-auth", auth_handler)

    auth = await refreshing_auth_client()
    api = google_nest_api.GoogleNestAPI(auth, PROJECT_ID)

    devices = await api.async_get_devices()
    assert len(devices) == 1
    assert devices[0].name == device_id
    assert devices[0].type == "sdm.devices.types.device-type1"


async def test_auth_refresh_error(
    app: aiohttp.web.Application,
    device_handler: DeviceHandler,
    refreshing_auth_client: Callable[[], Awaitable[AbstractAuth]],
) -> None:
    async def auth_handler(request: aiohttp.web.Request) -> aiohttp.web.Response:
        return aiohttp.web.Response(status=401)

    app.router.add_get("/refresh-auth", auth_handler)

    auth = await refreshing_auth_client()
    api = google_nest_api.GoogleNestAPI(auth, PROJECT_ID)
    with pytest.raises(AuthException):
        await api.async_get_devices()


async def test_no_devices(
    app: aiohttp.web.Application,
    recorder: Recorder,
    device_handler: DeviceHandler,
    api_client: Callable[[], Awaitable[google_nest_api.GoogleNestAPI]],
) -> None:
    api = await api_client()
    devices = await api.async_get_devices()
    assert devices == []


async def test_get_devices_missing_devices(
    app: aiohttp.web.Application,
    project_id: str,
    recorder: Recorder,
    api_client: Callable[[], Awaitable[google_nest_api.GoogleNestAPI]],
) -> None:
    reply_handler(app, f"/enterprises/{project_id}/devices", recorder, [{}])
    api = await api_client()
    devices = await api.async_get_devices()
    assert devices == []


async def test_get_device_missing_devices(
    app: aiohttp.web.Application,
    recorder: Recorder,
    api_client: Callable[[], Awaitable[google_nest_api.GoogleNestAPI]],
) -> None:
    app.router.add_get(
        "/enterprises/project-id1/devices/abc", NewHandler(recorder, [{}])
    )
    api = await api_client()
    device = await api.async_get_device("abc")
    assert device is None


async def test_no_structures(
    app: aiohttp.web.Application,
    structure_handler: StructureHandler,
    api_client: Callable[[], Awaitable[google_nest_api.GoogleNestAPI]],
) -> None:
    api = await api_client()
    structures = await api.async_get_structures()
    assert structures == []


async def test_get_structures_missing_structures(
    app: aiohttp.web.Application,
    project_id: str,
    recorder: Recorder,
    api_client: Callable[[], Awaitable[google_nest_api.GoogleNestAPI]],
) -> None:
    reply_handler(app, "/enterprises/{project_id}/structures", recorder, [{}])
    api = await api_client()
    structures = await api.async_get_structures()
    assert structures == []


async def test_get_structure_missing_structures(
    app: aiohttp.web.Application,
    recorder: Recorder,
    api_client: Callable[[], Awaitable[google_nest_api.GoogleNestAPI]],
) -> None:
    app.router.add_get(
        "/enterprises/project-id1/structures/abc", NewHandler(recorder, [{}])
    )
    api = await api_client()
    structure = await api.async_get_structure("abc")
    assert structure is None


@pytest.mark.parametrize(
    "status,error_dict,exc_type,err_match",
    [
        (
            HTTPStatus.BAD_REQUEST,
            {
                "code": 400,
                "message": "Some error message",
                "status": "FAILED_PRECONDITION",
            },
            ApiException,
            re.compile(r"Bad Request response from API \(400\).*Some error message"),
        ),
        (
            HTTPStatus.UNAUTHORIZED,
            {
                "code": 401,
                "message": "Some error message",
                "status": "UNAUTHENTICATED",
            },
            AuthException,
            re.compile(r"Unauthorized response from API \(401\).*Some error message"),
        ),
        (
            HTTPStatus.FORBIDDEN,
            {
                "code": 403,
                "message": "Some error message",
                "status": "PERMISSION_DENIED",
            },
            ApiForbiddenException,
            re.compile(r"Forbidden response from API \(403\):.*Some error message"),
        ),
        (
            HTTPStatus.NOT_FOUND,
            {
                "code": 404,
                "message": "Some error message",
                "status": "NOT_FOUND",
            },
            NotFoundException,
            re.compile(r"Not Found response from API \(404\):.*Some error message"),
        ),
        (
            HTTPStatus.INTERNAL_SERVER_ERROR,
            {
                "code": 500,
                "message": "Some error message",
                "status": "INTERNAL",
            },
            ApiException,
            re.compile(r"Internal Server Error response from API \(500\).*Some error message"),
        ),
        (
            503,
            {
                "code": 503,
                "message": "Some error message",
                "status": "UNAVAILABLE",
            },
            ApiException,
            re.escape("Service Unavailable response from API (503)"),
        ),
    ],
)
async def test_api_post_error_with_json_response(
    app: aiohttp.web.Application,
    device_handler: DeviceHandler,
    api_client: Callable[[], Awaitable[google_nest_api.GoogleNestAPI]],
    status: int,
    error_dict: dict,
    exc_type: type[Exception],
    err_match: str,
) -> None:
    device_id = device_handler.add_device(
        traits={
            "sdm.devices.traits.ThermostatTemperatureSetpoint": {
                "heatCelsius": 23.0,
                "coolCelsius": 24.0,
            },
        }
    )

    json_response = {
        "error": error_dict,
    }

    async def fail_handler(request: aiohttp.web.Request) -> aiohttp.web.Response:
        assert request.headers["Authorization"] == "Bearer %s" % FAKE_TOKEN
        return aiohttp.web.Response(
            status=status,
            body=json.dumps(json_response),
            content_type="application/json",
        )

    app.router.add_post(f"/{device_id}:executeCommand", fail_handler)

    api = await api_client()
    devices = await api.async_get_devices()
    assert len(devices) == 1
    device = devices[0]
    assert device.name == device_id
    trait = device.traits["sdm.devices.traits.ThermostatTemperatureSetpoint"]

    with pytest.raises(exc_type, match=err_match):
        await trait.set_heat(25.0)



"""Tests for the admin client library."""

from typing import Awaitable, Callable
from http import HTTPStatus

import aiohttp
import pytest

from google_nest_sdm.admin_client import AdminClient
from google_nest_sdm.auth import AbstractAuth
from google_nest_sdm.exceptions import (
    ApiException,
    ConfigurationException,
)

from .conftest import Recorder, NewHandler

GOOGLE_CLOUD_CONSOLE_PROJECT_ID = "google-cloud-console-project-id"
DEVICE_ACCESS_PROJECT_ID = "device-access-project-id"


@pytest.fixture(name="admin_client")
def mock_admin_client(
    auth_client: Callable[[], Awaitable[AbstractAuth]],
) -> Callable[[], Awaitable[AdminClient]]:

    async def _make_admin_client() -> AdminClient:
        mock_auth = await auth_client()
        return AdminClient(mock_auth, GOOGLE_CLOUD_CONSOLE_PROJECT_ID)

    return _make_admin_client


async def test_invalid_topic_format(
    app: aiohttp.web.Application,
    admin_client: Callable[[], Awaitable[AdminClient]],
    recorder: Recorder,
) -> None:
    """Test creating a topic."""
    client = await admin_client()
    with pytest.raises(ConfigurationException):
        await client.create_topic("some-topic")


async def test_create_topic(
    app: aiohttp.web.Application,
    admin_client: Callable[[], Awaitable[AdminClient]],
    recorder: Recorder,
) -> None:
    """Test creating a topic."""

    handler = NewHandler(
        recorder,
        [{}],
    )
    app.router.add_put(
        f"/projects/{GOOGLE_CLOUD_CONSOLE_PROJECT_ID}/topics/topic-name", handler
    )

    client = await admin_client()
    await client.create_topic(
        f"projects/{GOOGLE_CLOUD_CONSOLE_PROJECT_ID}/topics/topic-name"
    )

    assert recorder.request == {}


async def test_delete_topic(
    app: aiohttp.web.Application,
    admin_client: Callable[[], Awaitable[AdminClient]],
    recorder: Recorder,
) -> None:
    """Test deleting a topic."""

    handler = NewHandler(
        recorder,
        [{}],
    )
    app.router.add_delete(
        f"/projects/{GOOGLE_CLOUD_CONSOLE_PROJECT_ID}/topics/topic-name", handler
    )

    client = await admin_client()
    await client.delete_topic(
        f"projects/{GOOGLE_CLOUD_CONSOLE_PROJECT_ID}/topics/topic-name"
    )

    assert recorder.request == {}


async def test_list_topics(
    app: aiohttp.web.Application,
    admin_client: Callable[[], Awaitable[AdminClient]],
    recorder: Recorder,
) -> None:
    """Test listing topics."""

    handler = NewHandler(
        recorder,
        [
            {
                "topics": [
                    {"name": "projects/project-id/topics/topic1"},
                    {"name": "projects/project-id/topics/topic2"},
                ]
            }
        ],
    )
    app.router.add_get(f"/projects/{GOOGLE_CLOUD_CONSOLE_PROJECT_ID}/topics", handler)

    client = await admin_client()
    topics = await client.list_topics(f"projects/{GOOGLE_CLOUD_CONSOLE_PROJECT_ID}")

    assert topics == [
        "projects/project-id/topics/topic1",
        "projects/project-id/topics/topic2",
    ]
    assert recorder.request == {}


async def test_list_topics_invalid_prefix(
    app: aiohttp.web.Application,
    admin_client: Callable[[], Awaitable[AdminClient]],
    recorder: Recorder,
) -> None:
    """Test listing topics with an invalid prefix."""

    handler = NewHandler(
        recorder,
        [
            {
                "topics": [
                    {"name": "projects/project-id/topics/topic1"},
                    {"name": "projects/project-id/topics/topic2"},
                ]
            }
        ],
    )
    app.router.add_get(f"/projects/{GOOGLE_CLOUD_CONSOLE_PROJECT_ID}/topics", handler)

    client = await admin_client()
    with pytest.raises(ConfigurationException):
        await client.list_topics("projects/{GOOGLE_CLOUD_CONSOLE_PROJECT_ID}/topics")


async def test_get_topic(
    app: aiohttp.web.Application,
    admin_client: Callable[[], Awaitable[AdminClient]],
    recorder: Recorder,
) -> None:
    """Test getting a topic."""

    handler = NewHandler(
        recorder,
        [{"name": f"projects/{GOOGLE_CLOUD_CONSOLE_PROJECT_ID}/topics/topic-name"}],
    )
    app.router.add_get(
        f"/projects/{GOOGLE_CLOUD_CONSOLE_PROJECT_ID}/topics/topic-name", handler
    )

    client = await admin_client()
    response = await client.get_topic(
        f"projects/{GOOGLE_CLOUD_CONSOLE_PROJECT_ID}/topics/topic-name"
    )
    assert recorder.request == {}
    assert response == {
        "name": f"projects/{GOOGLE_CLOUD_CONSOLE_PROJECT_ID}/topics/topic-name"
    }


async def test_create_subscription(
    app: aiohttp.web.Application,
    admin_client: Callable[[], Awaitable[AdminClient]],
    recorder: Recorder,
) -> None:
    """Test creating a subscription."""

    handler = NewHandler(
        recorder,
        [{}],
    )
    app.router.add_put(
        f"/projects/{GOOGLE_CLOUD_CONSOLE_PROJECT_ID}/subscriptions/subscription-name",
        handler,
    )

    client = await admin_client()
    await client.create_subscription(
        f"projects/{GOOGLE_CLOUD_CONSOLE_PROJECT_ID}/topics/topic-name",
        f"projects/{GOOGLE_CLOUD_CONSOLE_PROJECT_ID}/subscriptions/subscription-name",
    )

    assert recorder.request == {
        "topic": f"projects/{GOOGLE_CLOUD_CONSOLE_PROJECT_ID}/topics/topic-name"
    }



async def test_create_subscription_failure(
    app: aiohttp.web.Application,
    admin_client: Callable[[], Awaitable[AdminClient]],
    recorder: Recorder,
) -> None:
    """Test creating a subscription."""

    handler = NewHandler(
        recorder,
        [{}],
        status=HTTPStatus.INTERNAL_SERVER_ERROR,
    )
    app.router.add_put(
        f"/projects/{GOOGLE_CLOUD_CONSOLE_PROJECT_ID}/subscriptions/subscription-name",
        handler,
    )

    client = await admin_client()
    with pytest.raises(ApiException, match="Internal Server Error response from API \(500\)"):
        await client.create_subscription(
            f"projects/{GOOGLE_CLOUD_CONSOLE_PROJECT_ID}/topics/topic-name",
            f"projects/{GOOGLE_CLOUD_CONSOLE_PROJECT_ID}/subscriptions/subscription-name",
        )


async def test_delete_subscription(
    app: aiohttp.web.Application,
    admin_client: Callable[[], Awaitable[AdminClient]],
    recorder: Recorder,
) -> None:
    """Test deleting a subscription."""

    handler = NewHandler(
        recorder,
        [{}],
    )
    app.router.add_delete(
        f"/projects/{GOOGLE_CLOUD_CONSOLE_PROJECT_ID}/subscriptions/subscription-name",
        handler,
    )

    client = await admin_client()
    await client.delete_subscription(
        f"projects/{GOOGLE_CLOUD_CONSOLE_PROJECT_ID}/subscriptions/subscription-name"
    )

    assert recorder.request == {}


async def test_list_subscriptions(
    app: aiohttp.web.Application,
    admin_client: Callable[[], Awaitable[AdminClient]],
    recorder: Recorder,
) -> None:
    """Test listing subscriptions."""

    handler = NewHandler(
        recorder,
        [
            {
                "subscriptions": [
                    {
                        "name": f"projects/{GOOGLE_CLOUD_CONSOLE_PROJECT_ID}/subscriptions/subscription1"
                    },
                    {
                        "name": f"projects/{GOOGLE_CLOUD_CONSOLE_PROJECT_ID}/subscriptions/subscription2"
                    },
                ]
            }
        ],
    )
    app.router.add_get(
        f"/projects/{GOOGLE_CLOUD_CONSOLE_PROJECT_ID}/subscriptions", handler
    )

    client = await admin_client()
    subscriptions = await client.list_subscriptions(
        f"projects/{GOOGLE_CLOUD_CONSOLE_PROJECT_ID}"
    )

    assert subscriptions == [
        {
            "name": "projects/google-cloud-console-project-id/subscriptions/subscription1"
        },
        {
            "name": "projects/google-cloud-console-project-id/subscriptions/subscription2"
        },
    ]


async def test_invalid_subscription_format(
    app: aiohttp.web.Application,
    admin_client: Callable[[], Awaitable[AdminClient]],
    recorder: Recorder,
) -> None:
    """Test creating a subscription."""
    client = await admin_client()
    with pytest.raises(ConfigurationException):
        await client.create_subscription(
            f"projects/{GOOGLE_CLOUD_CONSOLE_PROJECT_ID}/topics/topic-name",
            "some-subscription",
        )
    with pytest.raises(ConfigurationException):
        await client.create_subscription(
            "some-topic",
            f"projects/{GOOGLE_CLOUD_CONSOLE_PROJECT_ID}/subscriptions/subscription-name",
        )


async def test_list_eligible_topics(
    app: aiohttp.web.Application,
    admin_client: Callable[[], Awaitable[AdminClient]],
    recorder: Recorder,
) -> None:
    """Test listing eligible topics."""

    # SDM created pubsub topic exists (but is not visible, which is expected) exists
    sdm_handler = NewHandler(
        recorder,
        [
            {
                "error": {
                    "code": 403,
                    "message": "User not authorized to perform this action.",
                    "status": "PERMISSION_DENIED",
                }
            },
        ],
        status=HTTPStatus.FORBIDDEN,
    )
    app.router.add_get(
        f"/projects/sdm-prod/topics/enterprise-{DEVICE_ACCESS_PROJECT_ID}", sdm_handler
    )
    # Cloud topic also exists
    cloud_handler = NewHandler(
        recorder,
        [
            {
                "topics": [
                    {
                        "name": f"projects/{GOOGLE_CLOUD_CONSOLE_PROJECT_ID}/topics/sdm-testing"
                    }
                ]
            }
        ],
    )
    app.router.add_get(f"/projects/{GOOGLE_CLOUD_CONSOLE_PROJECT_ID}/topics", cloud_handler)

    client = await admin_client()
    eligible_topics = await client.list_eligible_topics(DEVICE_ACCESS_PROJECT_ID)
    assert eligible_topics.topic_names == [
        "projects/sdm-prod/topics/enterprise-device-access-project-id",
        "projects/google-cloud-console-project-id/topics/sdm-testing",
    ]


async def test_list_eligible_topics_no_sdm_topic(
    app: aiohttp.web.Application,
    admin_client: Callable[[], Awaitable[AdminClient]],
    recorder: Recorder,
) -> None:
    """Test listing eligible topics when no SDM topic exists."""

    # SDM created pubsub topic does not exist
    sdm_handler = NewHandler(
        recorder,
        [
            {
                "error": {
                    "code": 404,
                    "message": f"Resource not found (resource=enterprise-{DEVICE_ACCESS_PROJECT_ID}).",
                    "status": "NOT_FOUND",
                }
            }
        ],
        status=HTTPStatus.NOT_FOUND,
    )
    app.router.add_get(
        f"/projects/sdm-prod/topics/enterprise-{DEVICE_ACCESS_PROJECT_ID}", sdm_handler
    )

    # Cloud topic exists
    cloud_handler = NewHandler(
        recorder,
        [
            {
                "topics": [
                    {
                        "name": f"projects/{GOOGLE_CLOUD_CONSOLE_PROJECT_ID}/topics/sdm-testing"
                    }
                ]
            }
        ],
    )
    app.router.add_get(f"/projects/{GOOGLE_CLOUD_CONSOLE_PROJECT_ID}/topics", cloud_handler)

    client = await admin_client()
    eligible_topics = await client.list_eligible_topics(DEVICE_ACCESS_PROJECT_ID)
    assert eligible_topics.topic_names == [
        "projects/google-cloud-console-project-id/topics/sdm-testing"
    ]


@pytest.mark.parametrize(
    "sdm_status, cloud_status",
    [
        (HTTPStatus.INTERNAL_SERVER_ERROR, HTTPStatus.OK),
        (HTTPStatus.OK, HTTPStatus.INTERNAL_SERVER_ERROR),
    ],
)
async def test_list_cloud_console_api_error(
    app: aiohttp.web.Application,
    admin_client: Callable[[], Awaitable[AdminClient]],
    recorder: Recorder,
    sdm_status: HTTPStatus,
    cloud_status: HTTPStatus,
) -> None:
    """Test listing eligible topics when an error occurs listing the cloud console topics."""

    # SDM created pubsub topic exists (but is not visible, which is expected) exists
    sdm_handler = NewHandler(
        recorder,
        [
            {
                "error": {
                    "code": 403,
                    "message": "User not authorized to perform this action.",
                    "status": "PERMISSION_DENIED",
                }
            },
        ],
        status=sdm_status,
    )
    app.router.add_get(
        f"/projects/sdm-prod/topics/enterprise-{DEVICE_ACCESS_PROJECT_ID}", sdm_handler
    )
    # Cloud topic also exists
    cloud_handler = NewHandler(
        recorder,
        [{}],
        status=cloud_status,
    )
    app.router.add_get(f"/projects/{GOOGLE_CLOUD_CONSOLE_PROJECT_ID}", cloud_handler)

    client = await admin_client()
    with pytest.raises(ApiException):
        await client.list_eligible_topics(DEVICE_ACCESS_PROJECT_ID)


async def test_list_eligible_subscriptions(
    app: aiohttp.web.Application,
    admin_client: Callable[[], Awaitable[AdminClient]],
    recorder: Recorder,
) -> None:
    """Test listing eligible subscriptions."""

    cloud_handler = NewHandler(
        recorder,
        [
            {
                "subscriptions": [
                    {
                        "name": f"projects/{GOOGLE_CLOUD_CONSOLE_PROJECT_ID}/subscriptions/sdm-testing-sub",
                        "topic": f"projects/{GOOGLE_CLOUD_CONSOLE_PROJECT_ID}/topics/sdm-testing",
                        "pushConfig": {},
                        "ackDeadlineSeconds": 10,
                        "messageRetentionDuration": "604800s",
                        "expirationPolicy": {"ttl": "2678400s"},
                        "state": "ACTIVE",
                    }
                ]
            }
        ],
    )
    app.router.add_get(
        f"/projects/{GOOGLE_CLOUD_CONSOLE_PROJECT_ID}/subscriptions", cloud_handler
    )

    client = await admin_client()
    eligible_subscriptions = await client.list_eligible_subscriptions(
        expected_topic_name=f"projects/{GOOGLE_CLOUD_CONSOLE_PROJECT_ID}/topics/sdm-testing"
    )
    assert eligible_subscriptions.subscription_names == [
        f"projects/{GOOGLE_CLOUD_CONSOLE_PROJECT_ID}/subscriptions/sdm-testing-sub",
    ]



"""A device from the Smart Device Management API."""

from __future__ import annotations

import datetime
import logging
from typing import Any, Awaitable, Callable
from dataclasses import dataclass, field, fields, asdict

from mashumaro import field_options, DataClassDictMixin
from mashumaro.config import BaseConfig
from mashumaro.types import SerializationStrategy

from . import camera_traits, device_traits, doorbell_traits, thermostat_traits
from .auth import AbstractAuth
from .doorbell_traits import DoorbellChimeTrait
from .diagnostics import Diagnostics, redact_data
from .event import EventMessage, EventProcessingError
from .event_media import EventMediaManager
from .traits import Command
from .model import TraitDataClass, SDM_PREFIX, TRAITS

_LOGGER = logging.getLogger(__name__)


@dataclass
class ParentRelation(DataClassDictMixin):
    """Represents the parent structure/room of the current resource."""

    parent: str
    display_name: str = field(metadata=field_options(alias="displayName"))

    class Config(BaseConfig):
        serialize_by_alias = True


@dataclass
class TraitTypes(TraitDataClass):
    """Data model for parsing traits in the Google Nest SDM API."""

    # Device Traits
    connectivity: device_traits.ConnectivityTrait | None = field(
        metadata=field_options(
            alias="sdm.devices.traits.Connectivity",
        ),
        default=None,
    )
    fan: device_traits.FanTrait | None = field(
        metadata=field_options(
            alias="sdm.devices.traits.Fan",
        ),
        default=None,
    )
    info: device_traits.InfoTrait | None = field(
        metadata=field_options(
            alias="sdm.devices.traits.Info",
        ),
        default=None,
    )
    humidity: device_traits.HumidityTrait | None = field(
        metadata=field_options(
            alias="sdm.devices.traits.Humidity",
        ),
        default=None,
    )
    temperature: device_traits.TemperatureTrait | None = field(
        metadata=field_options(
            alias="sdm.devices.traits.Temperature",
        ),
        default=None,
    )

    # Thermostat Traits
    thermostat_eco: thermostat_traits.ThermostatEcoTrait | None = field(
        metadata=field_options(
            alias="sdm.devices.traits.ThermostatEco",
        ),
        default=None,
    )
    thermostat_hvac: thermostat_traits.ThermostatHvacTrait | None = field(
        metadata=field_options(
            alias="sdm.devices.traits.ThermostatHvac",
        ),
        default=None,
    )
    thermostat_mode: thermostat_traits.ThermostatModeTrait | None = field(
        metadata=field_options(
            alias="sdm.devices.traits.ThermostatMode",
        ),
        default=None,
    )
    thermostat_temperature_setpoint: (
        thermostat_traits.ThermostatTemperatureSetpointTrait | None
    ) = field(
        metadata=field_options(
            alias="sdm.devices.traits.ThermostatTemperatureSetpoint",
        ),
        default=None,
    )

    # # Camera Traits
    camera_image: camera_traits.CameraImageTrait | None = field(
        metadata=field_options(
            alias="sdm.devices.traits.CameraImage",
        ),
        default=None,
    )
    camera_live_stream: camera_traits.CameraLiveStreamTrait | None = field(
        metadata=field_options(alias="sdm.devices.traits.CameraLiveStream"),
        default=None,
    )
    camera_event_image: camera_traits.CameraEventImageTrait | None = field(
        metadata=field_options(
            alias="sdm.devices.traits.CameraEventImage",
        ),
        default=None,
    )
    camera_motion: camera_traits.CameraMotionTrait | None = field(
        metadata=field_options(
            alias="sdm.devices.traits.CameraMotion",
        ),
        default=None,
    )
    camera_person: camera_traits.CameraPersonTrait | None = field(
        metadata=field_options(
            alias="sdm.devices.traits.CameraPerson",
        ),
        default=None,
    )
    camera_sound: camera_traits.CameraSoundTrait | None = field(
        metadata=field_options(
            alias="sdm.devices.traits.CameraSound",
        ),
        default=None,
    )
    camera_clip_preview: camera_traits.CameraClipPreviewTrait | None = field(
        metadata=field_options(
            alias="sdm.devices.traits.CameraClipPreview",
        ),
        default=None,
    )

    # # Doorbell Traits
    doorbell_chime: doorbell_traits.DoorbellChimeTrait | None = field(
        metadata=field_options(
            alias="sdm.devices.traits.DoorbellChime",
        ),
        default=None,
    )


class ParentRelationsSerializationStrategy(SerializationStrategy, use_annotations=True):
    """Parser to ignore invalid parent relations."""

    def serialize(self, value: list[ParentRelation]) -> list[dict[str, Any]]:
        return [x.to_dict() for x in value]

    def deserialize(self, value: list[dict[str, Any]]) -> list[ParentRelation]:
        return [
            ParentRelation.from_dict(relation)
            for relation in value
            if "parent" in relation and "displayName" in relation
        ]


def _name_required() -> str:
    """Raise an error if the name field is not provided.

    This is a workaround for the fact that dataclasses children can't have
    default fields out of order from the subclass.
    """
    raise ValueError("Field 'name' is required")


@dataclass
class Device(TraitTypes):
    """Class that represents a device object in the Google Nest SDM API."""

    name: str = field(default_factory=_name_required)
    """Resource name of the device such as 'enterprises/XYZ/devices/123'."""

    type: str | None = None
    """Type of device for display purposes.

    The device type should not be used to deduce or infer functionality of
    the actual device it is assigned to. Instead, use the returned traits for
    the device.
    """

    relations: list[ParentRelation] = field(
        metadata=field_options(alias="parentRelations"), default_factory=list
    )
    """Represents the parent structure or room of the device."""

    _auth: AbstractAuth = field(init=False, metadata={"serialize": "omit"})
    _diagnostics: Diagnostics = field(init=False, metadata={"serialize": "omit"})
    _event_media_manager: EventMediaManager = field(
        init=False, metadata={"serialize": "omit"}
    )
    _callbacks: list[Callable[[EventMessage], Awaitable[None]]] = field(
        init=False, metadata={"serialize": "omit"}, default_factory=list
    )
    _trait_event_ts: dict[str, datetime.datetime] = field(
        init=False, metadata={"serialize": "omit"}, default_factory=dict
    )

    @staticmethod
    def MakeDevice(raw_data: dict[str, Any], auth: AbstractAuth) -> Device:
        """Create a device with the appropriate traits."""

        # Hack for incorrect nest API response values
        if (type := raw_data.get("type")) and type == "sdm.devices.types.DOORBELL":
            if TRAITS not in raw_data:
                raw_data[TRAITS] = {}
            raw_data[TRAITS][DoorbellChimeTrait.NAME] = {}

        device: Device = Device.parse_trait_object(raw_data)
        device._auth = auth
        device._diagnostics = Diagnostics()
        cmd = Command(raw_data["name"], auth, device._diagnostics.subkey("command"))
        for trait in device.traits.values():
            if hasattr(trait, "_cmd"):
                trait._cmd = cmd

        event_traits = {
            trait.EVENT_NAME
            for trait in device.traits.values()
            if hasattr(trait, "EVENT_NAME")
        }
        device._event_media_manager = EventMediaManager(
            device.name or "",
            device.traits,
            event_traits,
            diagnostics=device._diagnostics.subkey("event_media"),
        )
        return device

    def add_update_listener(self, target: Callable[[], None]) -> Callable[[], None]:
        """Register a simple event listener notified on updates.

        This will not block on media being fetched. To wait for media, use
        the callback form the `EventMediaManager`.

        The return value is a callable that will unregister the callback.
        """

        async def handle_event(event_message: EventMessage) -> None:
            target()

        return self.add_event_callback(handle_event)

    def add_event_callback(
        self, target: Callable[[EventMessage], Awaitable[None]]
    ) -> Callable[[], None]:
        """Register an event callback for updates to this device.

        This will not block on media being fetched. To wait for media, use
        the callback form the `EventMediaManager`.

        The return value is a callable that will unregister the callback.
        """
        self._callbacks.append(target)

        def remove_callback() -> None:
            """Remove the event_callback."""
            self._callbacks.remove(target)

        return remove_callback

    async def async_handle_event(self, event_message: EventMessage) -> None:
        """Process an event from the pubsub subscriber.
        
        This will invoke any directly registered callbacks (before fetching media)
        as well as any callbacks registered with the event media manager that
        fire post-media.
        """
        _LOGGER.debug(
            "Processing update %s @ %s", event_message.event_id, event_message.timestamp
        )
        if not event_message.resource_update_name:
            raise EventProcessingError("Event was not resource update event")
        if self.name != event_message.resource_update_name:
            raise EventProcessingError(
                f"Mismatch {self.name} != {event_message.resource_update_name}"
            )
        self._async_handle_traits(event_message)
        for callback in self._callbacks:
            await callback(event_message)
        await self._event_media_manager.async_handle_events(event_message)

    def _async_handle_traits(self, event_message: EventMessage) -> None:
        traits = event_message.resource_update_traits
        if not traits:
            return
        _LOGGER.debug("Trait update %s", traits)
        # Parse the traits using a separate object, then overwrite
        # each present field with an updated copy of the original trait with
        # the new fields merged in.
        parsed_traits = TraitTypes.parse_trait_object({TRAITS: traits})
        for trait_field in fields(parsed_traits):
            if (
                (alias := trait_field.metadata.get("alias")) is None
                or not alias.startswith(SDM_PREFIX)
                or not (new := getattr(parsed_traits, trait_field.name))
            ):
                continue
            # Discard updates to traits that are newer than the update
            if (
                self._trait_event_ts
                and (ts := self._trait_event_ts.get(trait_field.name))
                and ts > event_message.timestamp
            ):
                _LOGGER.debug("Discarding stale update (%s)", event_message.timestamp)
                continue

            # Only merge updates into existing models, updating the existing
            # fields present in the update trait
            if not (existing := getattr(self, trait_field.name)):
                continue
            for k, v in asdict(new).items():
                if v is not None:
                    setattr(existing, k, v)
            self._trait_event_ts[trait_field.name] = event_message.timestamp

    @property
    def event_media_manager(self) -> EventMediaManager:
        return self._event_media_manager

    @property
    def parent_relations(self) -> dict:
        """Room or structure for the device."""
        return {relation.parent: relation.display_name for relation in self.relations}

    def delete_relation(self, parent: str) -> None:
        """Remove a device relationship with the parent."""
        self.relations = [
            relation for relation in self.relations if relation.parent != parent
        ]

    def create_relation(self, relation: ParentRelation) -> None:
        """Add a new device relation."""
        self.relations.append(relation)

    def get_diagnostics(self) -> dict[str, Any]:
        return {
            "data": redact_data(self.raw_data),
            **self._diagnostics.as_dict(),
        }

    class Config(TraitTypes.Config):
        serialization_strategy = {
            list[ParentRelation]: ParentRelationsSerializationStrategy(),
        }



"""Tests for doorbell traits."""

from typing import Any, Callable, Dict

from google_nest_sdm.device import Device


def test_doorbell_chime(fake_device: Callable[[Dict[str, Any]], Device]) -> None:
    device = fake_device(
        {
            "name": "my/device/name",
            "traits": {
                "sdm.devices.traits.DoorbellChime": {},
            },
        }
    )
    assert device.traits.keys() == {"sdm.devices.traits.DoorbellChime"}


def test_doorbell_chime_trait_hack(
    fake_device: Callable[[Dict[str, Any]], Device]
) -> None:
    """Adds the DoorbellChime trait even when missing from the API to fix an API bug."""
    device = fake_device(
        {
            "name": "my/device/name",
            "type": "sdm.devices.types.DOORBELL",
            "traits": {},
        }
    )
    assert device.type == "sdm.devices.types.DOORBELL"
    assert device.traits.keys() == {"sdm.devices.traits.DoorbellChime"}


def test_doorbell_chime_trait_hack_empty_traits(
    fake_device: Callable[[Dict[str, Any]], Device]
) -> None:
    """Adds the DoorbellChime trait even when missing from the API to fix an API bug."""
    device = fake_device(
        {
            "name": "my/device/name",
            "type": "sdm.devices.types.DOORBELL",
        }
    )
    assert device.type == "sdm.devices.types.DOORBELL"
    assert device.traits.keys() == {"sdm.devices.traits.DoorbellChime"}


def test_doorbell_chime_trait_hack_not_applied(
    fake_device: Callable[[Dict[str, Any]], Device]
) -> None:
    """The doorbell chime trait hack is not applied for other types."""
    device = fake_device(
        {
            "name": "my/device/name",
            "type": "sdm.devices.types.CAMERA",
            "traits": {},
        }
    )
    assert device.type == "sdm.devices.types.CAMERA"
    assert device.traits.keys() == set()
